#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename:        $HeadURL$
#
#  Version:         $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .secure.GXDLMSSecureClient import GXDLMSSecureClient
from .enums import InterfaceType, Command, DataType
from .GXDLMS import GXDLMS
from .internal._GXCommon import _GXCommon
from .GXByteBuffer import GXByteBuffer
from .GXDLMSLNParameters import GXDLMSLNParameters
from .GXDLMSXmlSettings import GXDLMSXmlSettings
from .GXDLMSTranslator import GXDLMSTranslator
from .enums.Security import Security
from .GXDLMSXmlPdu import GXDLMSXmlPdu

class GXDLMSXmlClient(GXDLMSSecureClient):
    """
    GXDLMS Xml client implements methods to communicate with DLMS/COSEM metering
    devices using XML.
    """

    #
    # Constructor
    #
    # @param type
    #            XML type.
    #
    def __init__(self, type_):
        GXDLMSSecureClient.__init__(self)
        # XML client don't throw exceptions. It serializes them as a default. Set
        # value to true, if exceptions are thrown.
        #
        self.throwExceptions = False
        self.translator = GXDLMSTranslator(type_)
        self.translator.hex = False
        self.useLogicalNameReferencing = True


    @classmethod
    def removeRecursively(cls, node, nodeType, name):
        if node.getNodeType() == nodeType and (name is None or node.getNodeName() == name):
            node.getParentNode().removeChild(node)
        else:
            #  check the children recursively
            list_ = node.getChildNodes()
            i = 0
            while i < list_.getLength():
                cls.removeRecursively(list_.item(i), nodeType, name)
                i += 1

    #
    # Load XML commands from the string or file.
    #
    # @param xml
    #            XML string
    # @param settings
    #            Load settings.
    # Loaded XML objects.
    # pylint: disable=too-many-locals,too-many-nested-blocks
    def load(self, doc, loadSettings):
        #  Remove comments.
        self.removeRecursively(doc, Node.COMMENT_NODE, None)
        actions = list()
        description = None
        error = None
        errorUrl = None
        sleep = None
        pos = 0
        while pos != doc.getChildNodes().getLength():
            m1 = doc.getChildNodes().item(pos)
            if m1.getNodeType() == Node.ELEMENT_NODE:
                pos2 = 0
                while pos2 != m1.getChildNodes().getLength():
                    node = m1.getChildNodes().item(pos2)
                    if node.getNodeType() == Node.ELEMENT_NODE:
                        if node.getNodeName() == "Description":
                            description = node.getNodeValue()
                            pos2 += 1
                            continue
                        if node.getNodeName() == "Error":
                            error = node.getNodeValue()
                            pos2 += 1
                            continue
                        if node.getNodeName() == "ErrorUrl":
                            errorUrl = node.getNodeValue()
                            pos2 += 1
                            continue
                        if node.getNodeName() == "Sleep":
                            sleep = node.getNodeValue()
                            pos2 += 1
                            continue
                        if loadSettings and node.getNodeName() == "GetRequest":
                            if loadSettings.start() and loadSettings.end:
                                pos3 = 0
                                while pos3 != node.getChildNodes().getLength():
                                    n1 = node.getChildNodes().item(pos3)
                                    if n1.getNodeName() == "GetRequestNormal":
                                        pos4 = 0
                                        while pos4 != n1.getChildNodes().getLength():
                                            n2 = n1.getChildNodes().item(pos4)
                                            if n2.getNodeName() == "AccessSelection":
                                                pos5 = 0
                                                while pos5 != n2.getChildNodes().getLength():
                                                    n3 = n2.getChildNodes().item(pos5)
                                                    if n3.getNodeName() == "AccessSelector":
                                                        if not n3.getAttributes().getNamedItem("Value").getNodeValue() == "1":
                                                            break
                                                    elif n3.getNodeName() == "AccessParameters":
                                                        pos6 = 0
                                                        while pos6 != n3.getChildNodes().getLength():
                                                            n4 = n3.getChildNodes().item(pos6)
                                                            if n4.getNodeName() == "Structure":
                                                                start = True
                                                                pos7 = 0
                                                                while pos7 != n4.getChildNodes().getLength():
                                                                    n5 = n4.getChildNodes().item(pos7)
                                                                    if n5.getNodeName() == "OctetString":
                                                                        bb = GXByteBuffer()
                                                                        if start:
                                                                            _GXCommon.setData(self.settings, bb, DataType.OCTET_STRING, loadSettings.getStart())
                                                                            n5.getAttributes().getNamedItem("Value").setNodeValue(bb.toHex(False, 2))
                                                                            start = False
                                                                        else:
                                                                            _GXCommon.setData(self.settings, bb, DataType.OCTET_STRING, loadSettings.getEnd())
                                                                            n5.getAttributes().getNamedItem("Value").setNodeValue(bb.toHex(False, 2))
                                                                            break
                                                                    pos7 += 1
                                                                break
                                                            pos6 += 1
                                                        break
                                                    pos5 += 1
                                                break
                                            pos4 += 1
                                        break
                                    pos3 += 1
                        s = GXDLMSXmlSettings(self.translator.outputType, self.translator.hex, self.translator.showStringAsHex, self.translator.tagsByName)
                        s.settings.clientAddress = self.settings.clientAddress
                        s.settings.serverAddress = self.settings.serverAddress
                        reply = []
                        reply = self.translator.xmlToPdu(GXDLMSXmlPdu.getOuterXml(node), s)
                        if s.command == Command.SNRM and not s.settings.isServer:
                            self.settings.hdlc.maxInfoTX = s.settings.hdlc.maxInfoTX
                            self.settings.hdlc.maxInfoRX = s.settings.hdlc.maxInfoRX
                            self.settings.hdlc.windowSizeRX = s.settings.hdlc.windowSizeRX
                            self.settings.hdlc.windowSizeTX = s.settings.hdlc.windowSizeTX
                        elif s.command == Command.UA and s.settings.isServer:
                            self.settings.hdlc.maxInfoTX = s.settings.hdlc.maxInfoTX
                            self.settings.hdlc.maxInfoRX = s.settings.hdlc.maxInfoRX
                            self.settings.hdlc.windowSizeRX = s.settings.hdlc.windowSizeRX
                            self.settings.hdlc.windowSizeTX = s.settings.hdlc.windowSizeTX
                        if s.template:
                            reply = None
                        p = GXDLMSXmlPdu(s.command, node, reply)
                        if description:
                            p.description = description
                        if error:
                            p.error = error
                        if errorUrl:
                            p.errorUrl = errorUrl
                        if sleep:
                            p.sleep = int(sleep)
                        actions.append(p)
                    pos2 += 1
            pos += 1
        return actions

    def pduToMessages(self, pdu):
        messages = list()
        if pdu.command == Command.SNRM:
            messages.append(pdu.data)
        elif pdu.command == Command.UA:
            messages.append(pdu.data)
        elif pdu.command == Command.DISCONNECT_REQUEST:
            messages.append(GXDLMS.getHdlcFrame(self.settings, int(Command.DISCONNECT_REQUEST), GXByteBuffer(pdu.data)))
        else:
            reply = None
            if self.settings.interfaceType == InterfaceType.WRAPPER:
                if self.ciphering.security != Security.NONE:
                    p = GXDLMSLNParameters(self.settings, 0, pdu.command, 0x0, None, None, 0xff)
                    reply = GXByteBuffer(GXDLMS.cipher0(p, pdu.data))
                else:
                    reply = GXByteBuffer(pdu.data)
            else:
                if self.ciphering.security != Security.NONE:
                    p = GXDLMSLNParameters(self.settings, 0, pdu.command, 0x0, None, None, 0xff)
                    tmp = GXDLMS.cipher0(p, pdu.data)
                    reply = GXByteBuffer(len(tmp))
                    reply.set(_GXCommon.LLC_SEND_BYTES)
                    reply.set(tmp)
                else:
                    reply = GXByteBuffer(len(pdu.data))
                    reply.set(_GXCommon.LLC_SEND_BYTES)
                    reply.set(pdu.data)
            frame_ = 0
            while reply.position != len(reply):
                if self.settings.interfaceType == InterfaceType.WRAPPER:
                    messages.append(GXDLMS.getWrapperFrame(self.settings, pdu.command, reply))
                elif self.settings.interfaceType == InterfaceType.HDLC or self.settings.interfaceType == InterfaceType.HDLC_WITH_MODE_E:
                    if pdu.command == Command.AARQ:
                        frame_ = 0x10
                    elif pdu.command == Command.AARE:
                        frame_ = 0x30
                    elif pdu.command == Command.EVENT_NOTIFICATION:
                        frame_ = 0x13
                    messages.append(GXDLMS.getHdlcFrame(self.settings, frame_, reply))
                    if reply.position != len(reply):
                        frame_ = self.settings.getNextSend(False)
                elif self.settings.interfaceType == InterfaceType.PDU:
                    messages.append(reply.array())
                    break
                else:
                    raise ValueError("InterfaceType")
        return messages



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .enums import ErrorCode, DataType
from .GXDLMSSettings import GXDLMSSettings

# pylint: disable=bad-option-value,old-style-class,too-few-public-methods,too-many-instance-attributes
class ValueEventArgs:
    #
    # Constructor.
    #
    # @param s
    #            DLMS settings.
    # @param eventTarget
    #            Event target.
    # @param eventIndex
    #            Event index.
    # @param readSelector
    #            Optional read event selector.
    # @param forParameters
    #            Optional parameters.
    # pylint: disable=too-many-arguments
    def __init__(self, s, eventTarget, eventIndex, readSelector=0, forParameters=None):
        if isinstance(s, GXDLMSSettings):
            self.settings = s
        else:
            self.settings = s.settings
        # Data type of the value.
        self.dataType = DataType.NONE
        # Target DLMS object
        self.target = eventTarget
        # Attribute index.
        self.index = eventIndex
        # Optional selector.
        self.selector = readSelector
        # Optional parameters.
        self.parameters = forParameters
        # Object value.
        self.eventValue = None
        # Is request handled.
        self.handled = False
        # Occurred error.
        self.error = ErrorCode.OK
        # Is action.  This is reserved for internal use.
        self.action = False
        # Is value max PDU size skipped when converting data to bytes.
        self.skipMaxPduSize = False
        # Is reply handled as byte array or octect string.
        self.byteArray = False
        # Row to PDU is used with Profile Generic to tell how many rows are fit
        # to
        # one PDU.
        self.rowToPdu = 0
        # Rows begin index.
        self.rowBeginIndex = 0
        # Rows end index.
        self.rowEndIndex = 0
        # DLMS server.
        self.server = None
        # Invoke ID.
        self.invokeId = 0
        self.rowEndIndex = 0
        self.rowBeginIndex = 0



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from abc import ABCMeta, abstractmethod
import datetime
from .GXReplyData import GXReplyData
from .GXDLMSSettings import GXDLMSSettings
from .enums.Command import Command
from .GXDLMSLongTransaction import GXDLMSLongTransaction
from .ServiceError import ServiceError
from .enums.Service import Service
from .ConfirmedServiceError import ConfirmedServiceError
from .internal._GXCommon import _GXCommon
from .enums.InterfaceType import InterfaceType
from .GXDLMS import GXDLMS
from .GXDLMSSNCommandHandler import GXDLMSSNCommandHandler
from .ConnectionState import ConnectionState
from .GXDLMSLNParameters import GXDLMSLNParameters
from .GXDLMSSNParameters import GXDLMSSNParameters
from .GXByteBuffer import GXByteBuffer
from .enums.ErrorCode import ErrorCode
from .GXDLMSConfirmedServiceError import GXDLMSConfirmedServiceError
from .enums.RequestTypes import RequestTypes
from .enums.Authentication import Authentication
from ._HDLCInfo import _HDLCInfo
from .GXSecure import GXSecure
from ._GXAPDU import _GXAPDU
from .GXDLMSException import GXDLMSException
from .enums.AssociationResult import AssociationResult
from .objects.enums.AssociationStatus import AssociationStatus
from .enums.SourceDiagnostic import SourceDiagnostic
from .objects.enums.ApplicationContextName import ApplicationContextName
from .enums.ObjectType import ObjectType
from .ValueEventArgs import ValueEventArgs
from .enums.DataType import DataType
from .objects.GXDLMSAssociationLogicalName import GXDLMSAssociationLogicalName
from .objects.GXDLMSAssociationShortName import GXDLMSAssociationShortName
from .GXDateTime import GXDateTime
from .enums.Conformance import Conformance
from .objects.IGXDLMSBase import IGXDLMSBase
from .enums.Initiate import Initiate
from .enums.Security import Security
from .GXDLMSLNCommandHandler import GXDLMSLNCommandHandler
from .GXHdlcSettings import GXHdlcSettings

# pylint:disable=too-many-public-methods,too-many-instance-attributes,useless-object-inheritance
class GXDLMSServer(object):
    __metaclass__ = ABCMeta
    #
    # Constructor.
    # logicalNameReferencing: Is logical name referencing used.
    # interfaceType: Interface type.
    #
    def __init__(self, logicalNameReferencing, interfaceType):
        self.info = GXReplyData()
        # Received data.
        self.receivedData = GXByteBuffer()
        # Reply data.
        self.replyData = GXByteBuffer()
        # Long get or read transaction information.
        self.transaction = None
        # Server settings.
        self.settings = GXDLMSSettings(True)
        # Is server initialized.
        self.initialized = False
        # When data was received last time.
        self.dataReceived = 0
        self.settings.setUseLogicalNameReferencing(logicalNameReferencing)
        self.settings.interfaceType = interfaceType
        self.hdlc = None
        self.wrapper = None
        self.reset()

    def getItems(self):
        return self.settings.objects

    # List of objects that meter supports.
    items = property(getItems)

    def getGbtWindowSize(self):
        return self.settings.gbtWindowSize

    def setGbtWindowSize(self, value):
        self.settings.gbtWindowSize = value

    #GBT Window size.
    gbtWindowSize = property(getGbtWindowSize, setGbtWindowSize)

    def getPushClientAddress(self):
        return self.settings.pushClientAddress

    def setPushClientAddress(self, value):
        self.settings.pushClientAddress = value

    #client address for push messages.
    pushClientAddress = property(getPushClientAddress, setPushClientAddress)
 #
    # Information from the connection size that server can
    #      handle.
    #
    @property
    def limits(self):
        """Obsolete. Use hdlcSettings instead."""
        return self.settings.hdlc

    #
    # HDLC framing settings.
    #
    @property
    def hdlcSettings(self):
        return self.settings.hdlc

    def getMaxReceivePDUSize(self):
        return self.maxReceivePDUSize

    def setMaxReceivePDUSize(self, value):
        self.maxReceivePDUSize = value

    #
    # Retrieves the maximum size of received PDU.  PDU size tells maximum size
    # of PDU packet.  Value can be from 0 to 0xFFFF.  By default the value is
    # 0xFFFF.
    #
    # Maximum size of received PDU.
    #
    maxReceivePDUSize = property(getMaxReceivePDUSize, setMaxReceivePDUSize)

    def getUseLogicalNameReferencing(self):
        return self.settings.getUseLogicalNameReferencing()

    def setUseLogicalNameReferencing(self, value):
        self.settings.setUseLogicalNameReferencing(value)

    #
    # Determines, whether Logical, or Short name, referencing is used.
    # Referencing depends on the device to communicate with.  Normally, a
    # device
    # supports only either Logical or Short name referencing.  The referencing
    # is defined by the device manufacturer.  If the referencing is wrong, the
    # SNMR message will fail.
    #
    # @see #getMaxReceivePDUSize
    # Is logical name referencing used.
    #
    useLogicalNameReferencing = property(getUseLogicalNameReferencing, setUseLogicalNameReferencing)

    def getConformance(self):
        return self.settings.proposedConformance

    def setConformance(self, value):
        self.settings.proposedConformance = value

    #
    # What kind of services server is offering.
    #
    conformance = property(getConformance, setConformance)

    #
    # Check is data sent to this server.
    #
    # @param serverAddress
    #            Server address.
    # @param clientAddress
    #            Client address.
    # True, if data is sent to this server.
    #
    @abstractmethod
    def isTarget(self, serverAddress, clientAddress):
        raise ValueError("isTarget is called.")

    #
    # Check whether the authentication and password are correct.
    #
    # @param authentication
    #            Authentication level.
    # @param password
    #            Password.
    # Source diagnostic.
    #
    @abstractmethod
    def onValidateAuthentication(self, authentication, password):
        raise ValueError("isTarget is called.")

    #
    # Get selected value(s).  This is called when example profile generic
    # request current value.
    #
    # @param args
    #            Value event arguments.
    #
    @abstractmethod
    def onPreGet(self, args):
        raise ValueError("isTarget is called.")

    #
    # Get selected value(s).  This is called when example profile generic
    # request current value.
    #
    # @param args
    #            Value event arguments.
    #
    @abstractmethod
    def onPostGet(self, args):
        raise ValueError("isTarget is called.")

    #
    # Find object.
    #
    # @param objectType
    #            Object type.
    # @param sn
    #            Short Name.  In Logical name referencing this is not used.
    # @param ln
    #            Logical Name.  In Short Name referencing this is not used.
    # Found object or null if object is not found.
    #
    @abstractmethod
    def onFindObject(self, objectType, sn, ln):
        raise ValueError("isTarget is called.")

    #
    # Called before read is executed.
    #
    # @param args
    #            Handled read requests.
    #
    @abstractmethod
    def onPreRead(self, args):
        raise ValueError("isTarget is called.")

    #
    # Called after read is executed.
    #
    # @param args
    #            Handled read requests.
    #
    @abstractmethod
    def onPostRead(self, args):
        raise ValueError("isTarget is called.")

    #
    # Called before write is executed..
    #
    # @param args
    #            Handled write requests.
    #
    @abstractmethod
    def onPreWrite(self, args):
        raise ValueError("isTarget is called.")

    #
    # Called after write is executed.
    #
    # @param args
    #            Handled write requests.
    #
    @abstractmethod
    def onPostWrite(self, args):
        raise ValueError("isTarget is called.")

    #
    # Accepted connection is made for the server.  All initialization is done
    # here.
    #
    # @param connectionInfo
    #            Connection info.
    #
    @abstractmethod
    def onConnected(self, connectionInfo):
        raise ValueError("isTarget is called.")

    #
    # Client has try to made invalid connection.  Password is incorrect.
    #
    # @param connectionInfo
    #            Connection info.
    #
    @abstractmethod
    def onInvalidConnection(self, connectionInfo):
        raise ValueError("isTarget is called.")

    #
    # Server has close the connection.  All clean up is made here.
    #
    # @param connectionInfo
    #            Connection info.
    #
    @abstractmethod
    def onDisconnected(self, connectionInfo):
        raise ValueError("isTarget is called.")

    #
    # Get attribute access mode.
    #
    # @param arg
    #            Value event argument.
    # Access mode.
    #
    @abstractmethod
    def onGetAttributeAccess(self, arg):
        raise ValueError("isTarget is called.")

    #
    # Get method access mode.
    #
    # @param arg
    #            Value event argument.
    # Method access mode.
    #
    @abstractmethod
    def onGetMethodAccess(self, arg):
        raise ValueError("onGetMethodAccess is called.")

    #
    # Called before action is executed.
    #
    # @param args
    #            Handled action requests.
    #
    @abstractmethod
    def onPreAction(self, args):
        raise ValueError("onPreAction is called.")

    #
    # Called after action is executed.
    #
    # @param args
    #            Handled action requests.
    #
    @abstractmethod
    def onPostAction(self, args):
        raise ValueError("onPostAction is called.")

    #
    # Add value of COSEM object to byte buffer.  AddData method can be used
    # with
    # GetDataNotificationMessage -method.  DLMS specification do not specify
    # the
    # structure of Data-Notification body.  So each manufacture can sent
    # different data.
    #
    # @param obj
    #            COSEM object.
    # @param index
    #            Attribute index.
    # @param buff
    #            Byte buffer.
    #
    def addData(self, obj, index, buff):
        dt = None
        e = ValueEventArgs(self.settings, obj, index, 0, None)
        value = obj.getValue(self.settings, e)
        dt = obj.getDataType(index)
        if dt == DataType.NONE and value:
            dt = _GXCommon.getDLMSDataType(value)
        _GXCommon.setData(self.settings, buff, dt, value)

    #
    # Generates data notification message.
    #
    # @param time
    #            Date time.  Set Date(0) if not added.
    # @param data
    #            Notification body.
    # Generated data notification message(s).
    #
    def generateDataNotificationMessages(self, time, data):
        if self.useLogicalNameReferencing:
            p = GXDLMSLNParameters(self.settings, 0, Command.DATA_NOTIFICATION, 0, None, data, 0xff)
            if time is None:
                p.time = None
            else:
                p.time = GXDateTime(time)
            reply = GXDLMS.getLnMessages(p)
        else:
            p2 = GXDLMSSNParameters(self.settings, Command.DATA_NOTIFICATION, 1, 0, data, None)
            reply = GXDLMS.getSnMessages(p2)
        if (self.settings.negotiatedConformance & Conformance.GENERAL_BLOCK_TRANSFER) == 0 and len(reply) != 1:
            raise ValueError("Data is not fit to one PDU. Use general block transfer.")
        return reply

    def generatePushSetupMessages(self, date, push):
        if push is None:
            raise ValueError("push")
        buff = GXByteBuffer()
        buff.setUInt8(DataType.STRUCTURE)
        _GXCommon.setObjectCount(len(push.pushObjectList), buff)
        for k, v in push.pushObjectList:
            self.addData(k, v.attributeIndex, buff)
        return self.generateDataNotificationMessages(date, buff)

    def getStoCChallenge(self):
        return self.settings.stoCChallenge

    def setStoCChallenge(self, value):
        self.settings.useCustomChallenge = value is not None
        self.settings.stoCChallenge = value

    #
    # Server to Client custom challenge.  This is for debugging
    #      purposes.  Reset
    # custom challenge settings StoCChallenge to null.
    # @param value Server to Client challenge.
    #
    stoCChallenge = property(getStoCChallenge, setStoCChallenge)

    #
    # Interface type.
    #
    def getInterfaceType(self):
        return self.settings.interfaceType


    def getStartingPacketIndex(self):
        return self.settings.blockIndex

    def setStartingPacketIndex(self, value):
        self.settings.blockIndex = value

    #
    # Set starting packet index.  Default is One based, but some meters
    #      use Zero
    # based value.  Usually this is not used.
    # @param value Zero based starting index.
    #
    startingPacketIndex = property(getStartingPacketIndex, setStartingPacketIndex)

    def getInvokeID(self):
        return self.settings.invokeId

    def setInvokeID(self, value):
        self.settings.invokeID = value

    # Invoke ID.
    invokeID = property(getInvokeID, setInvokeID)

    def getServiceClass(self):
        return self.settings.serviceClass

    def setServiceClass(self, value):
        self.settings.serviceClass = value

    # Used service class.
    serviceClass = property(getServiceClass, setServiceClass)

    def getPriority(self):
        return self.settings.priority

    def setPriority(self, value):
        self.settings.priority = value

    # Used priority.
    priority = property(getPriority, setPriority)

    #
    # Initialize server.  This must call after server objects are set.
    #
    def initialize(self):
        associationObject = None
        self.initialized = True
        pos = 0
        while pos != len(self.settings.objects):
            it = self.settings.objects[pos]
            if not it.logicalName:
                raise ValueError("Invalid Logical Name.")
            it.start = self
            if isinstance(it, (GXDLMSAssociationShortName,)) and not self.useLogicalNameReferencing:
                if len(it.objectList) == 0:
                    it.objectList.append(self.items)
                associationObject = it
            elif isinstance(it, (GXDLMSAssociationLogicalName,)) and self.useLogicalNameReferencing:
                ln = it
                if len(ln.objectList) == 0:
                    ln.objectList.append(self.items)
                associationObject = it
                ln.xDLMSContextInfo.maxReceivePduSize = self.settings.maxServerPDUSize
                ln.xDLMSContextInfo.maxSendPduSize = self.settings.maxServerPDUSize
            elif not isinstance(it, IGXDLMSBase):
                self.settings.objects.remove(pos)
                pos -= 1
            pos += 1
        if not associationObject:
            if self.useLogicalNameReferencing:
                it = GXDLMSAssociationLogicalName()
                it.xDLMSContextInfo.maxReceivePduSize = self.settings.maxServerPDUSize
                it.xDLMSContextInfo.maxSendPduSize = self.settings.maxServerPDUSize
                self.items.append(it)
                it.objectList.append(self.items)
            else:
                it2 = GXDLMSAssociationShortName()
                self.items.append(it2)
                it2.objectList.append(self.items)
        if not self.useLogicalNameReferencing:
            self.updateShortNames(False)

    def updateShortNames(self, force):
        sn = 0xA0
        offset = [0]
        count = [0]
        for it in self.settings.objects:
            if not isinstance(it, (GXDLMSAssociationShortName, GXDLMSAssociationLogicalName)):
                if force or it.shortName == 0:
                    it.setShortName(sn)
                    GXDLMS.getActionInfo(it.objectType, offset, count)
                    if count[0] != 0:
                        sn += offset[0] + (8 * count[0])
                    else:
                        sn += 8 * it.getAttributeCount()
                else:
                    sn = it.shortName

    def handleAarqRequest(self, data, connectionInfo):
        # pylint: disable=too-many-nested-blocks
        result = AssociationResult.ACCEPTED
        error = None
        self.settings.setCtoSChallenge(None)
        if self.settings.cipher:
            self.settings.cipher.setDedicatedKey(None)
        if self.settings.interfaceType == InterfaceType.WRAPPER:
            self.reset(True)
        diagnostic = SourceDiagnostic.NO_REASON_GIVEN
        try:
            diagnostic = _GXAPDU.parsePDU(self.settings, self.settings.cipher, data, None)
            if self.settings.negotiatedConformance == Conformance.NONE:
                result = AssociationResult.PERMANENT_REJECTED
                diagnostic = SourceDiagnostic.NO_REASON_GIVEN
                error = GXByteBuffer()
                error.setUInt8(0xE)
                error.setUInt8(ConfirmedServiceError.INITIATE_ERROR)
                error.setUInt8(ServiceError.INITIATE)
                error.setUInt8(Initiate.INCOMPATIBLE_CONFORMANCE)
            elif self.settings.maxPduSize < 64:
                result = AssociationResult.PERMANENT_REJECTED
                diagnostic = SourceDiagnostic.NO_REASON_GIVEN
                error = GXByteBuffer()
                error.setUInt8(0xE)
                error.setUInt8(ConfirmedServiceError.INITIATE_ERROR)
                error.setUInt8(ServiceError.INITIATE)
                error.setUInt8(Initiate.PDU_SIZE_TOO_SHORT)
            elif self.settings.dlmsVersion != 6:
                self.settings.dlmsVersion = 6
                result = AssociationResult.PERMANENT_REJECTED
                diagnostic = SourceDiagnostic.NO_REASON_GIVEN
                error = GXByteBuffer()
                error.setUInt8(0xE)
                error.setUInt8(ConfirmedServiceError.INITIATE_ERROR)
                error.setUInt8(ServiceError.INITIATE)
                error.setUInt8(Initiate.DLMS_VERSION_TOO_LOW)
            elif diagnostic != SourceDiagnostic.NONE:
                result = AssociationResult.PERMANENT_REJECTED
                diagnostic = SourceDiagnostic.NOT_SUPPORTED
                self.onInvalidConnection(connectionInfo)
            else:
                diagnostic = self.onValidateAuthentication(self.settings.authentication, self.settings.password)
                if diagnostic != SourceDiagnostic.NONE:
                    result = AssociationResult.PERMANENT_REJECTED
                elif self.settings.authentication > Authentication.LOW:
                    result = AssociationResult.ACCEPTED
                    diagnostic = SourceDiagnostic.AUTHENTICATION_REQUIRED
                    if self.useLogicalNameReferencing:
                        ln = self.items.findByLN(ObjectType.ASSOCIATION_LOGICAL_NAME, "0.0.40.0.0.255")
                        if ln:
                            if self.settings.cipher is None or self.settings.cipher.security == Security.NONE:
                                ln.applicationContextName.contextId = ApplicationContextName.LOGICAL_NAME
                            else:
                                ln.applicationContextName.contextId = ApplicationContextName.LOGICAL_NAME_WITH_CIPHERING
                            ln.authenticationMechanismName.mechanismId = self.settings.authentication
                            ln.associationStatus = AssociationStatus.ASSOCIATION_PENDING
                else:
                    if self.useLogicalNameReferencing:
                        ln = self.items.findByLN(ObjectType.ASSOCIATION_LOGICAL_NAME, "0.0.40.0.0.255")
                        if ln:
                            if self.settings.cipher is None or self.settings.cipher.security == Security.NONE:
                                ln.applicationContextName.contextId = ApplicationContextName.LOGICAL_NAME
                            else:
                                ln.applicationContextName.contextId = ApplicationContextName.LOGICAL_NAME_WITH_CIPHERING
                            ln.authenticationMechanismName.mechanismId = self.settings.authentication
                            ln.associationStatus = AssociationStatus.ASSOCIATED
                    self.settings.connected = self.settings.connected | ConnectionState.DLMS
        except GXDLMSConfirmedServiceError as e:
            result = AssociationResult.PERMANENT_REJECTED
            diagnostic = SourceDiagnostic.NO_REASON_GIVEN
            error = GXByteBuffer()
            error.setUInt8(0xE)
            error.setUInt8(e.confirmedServiceError)
            error.setUInt8(e.serviceError)
            error.setUInt8(e.serviceErrorValue)
        except GXDLMSException as e:
            result = e.result
            diagnostic = e.diagnostic
        if self.settings.interfaceType == InterfaceType.HDLC or self.settings.interfaceType == InterfaceType.HDLC_WITH_MODE_E:
            self.replyData.set(_GXCommon.LLC_REPLY_BYTES)
        if self.settings.authentication > Authentication.LOW:
            self.settings.setStoCChallenge(GXSecure.generateChallenge())
        _GXAPDU.generateAARE(self.settings, self.replyData, result, diagnostic, self.settings.cipher, error, None)

    def handleReleaseRequest(self, data):
        # pylint: disable=unused-argument
        if self.settings.interfaceType == InterfaceType.HDLC or self.settings.interfaceType == InterfaceType.HDLC_WITH_MODE_E:
            self.replyData.set(0, _GXCommon.LLC_REPLY_BYTES)
        tmp = _GXAPDU.getUserInformation(self.settings, self.settings.cipher)
        self.replyData.setUInt8(0x63)
        self.replyData.setUInt8(int((len(tmp))))
        self.replyData.setUInt8(0x80)
        self.replyData.setUInt8(0x01)
        self.replyData.setUInt8(0x00)
        self.replyData.setUInt8(0xBE)
        self.replyData.setUInt8(1 + len(tmp))
        self.replyData.setUInt8(4)
        self.replyData.setUInt8(len(tmp))
        self.replyData.set(tmp)

    def handleSnrmRequest(self, data):
        if self.hdlc:
            self.settings.hdlc.update(self.hdlcSettings)
        else:
            self.settings.hdlc.maxInfoRX = GXHdlcSettings.DEFAULT_MAX_INFO_RX
            self.settings.hdlc.maxInfoTX = GXHdlcSettings.DEFAULT_MAX_INFO_TX
            self.settings.hdlc.windowSizeRX = GXHdlcSettings.DEFAULT_WINDOWS_SIZE_RX
            self.settings.hdlc.windowSizeTX = GXHdlcSettings.DEFAULT_WINDOWS_SIZE_TX
        GXDLMS.parseSnrmUaResponse(data, self.settings.hdlc)
        self.reset(True)
        self.replyData.setUInt8(0x81)
        self.replyData.setUInt8(0x80)
        self.replyData.setUInt8(0)
        if self.hdlc:
            #If client wants send larger HDLC frames what meter accepts.
            if self.settings.hdlc.maxInfoTX > self.hdlc.maximumInfoLengthReceive:
                self.settings.hdlc.maxInfoTX = self.hdlc.maximumInfoLengthReceive
            if self.settings.hdlc.maxInfoRX > self.hdlc.maximumInfoLengthTransmit:
                self.settings.hdlc.maxInfoRX = self.hdlc.maximumInfoLengthTransmit
            if self.settings.hdlc.maxInfoRX > self.hdlc.maximumInfoLengthTransmit:
                self.settings.hdlc.windowSizeTX = self.hdlc.windowSizeReceive
            if self.settings.hdlc.maxInfoRX > self.hdlc.maximumInfoLengthTransmit:
                self.settings.hdlc.windowSizeRX = self.hdlc.windowSizeTransmit
        self.replyData.setUInt8(_HDLCInfo.MAX_INFO_TX)
        GXDLMS.appendHdlcParameter(self.replyData, self.hdlc.maxInfoTX)
        self.replyData.setUInt8(_HDLCInfo.MAX_INFO_RX)
        GXDLMS.appendHdlcParameter(self.replyData, self.hdlc.maxInfoRX)
        self.replyData.setUInt8(_HDLCInfo.WINDOW_SIZE_TX)
        self.replyData.setUInt8(4)
        self.replyData.setUInt32(self.hdlc.windowSizeTX)
        self.replyData.setUInt8(_HDLCInfo.WINDOW_SIZE_RX)
        self.replyData.setUInt8(4)
        self.replyData.setUInt32(self.hdlc.windowSizeRX)
        self.replyData.setUInt8(2, len(self.replyData) - 3)
        self.settings.connected = ConnectionState.HDLC

    def generateDisconnectRequest(self):
        self.replyData.setUInt8(0x81)
        self.replyData.setUInt8(0x80)
        self.replyData.setUInt8(0)
        self.replyData.setUInt8(_HDLCInfo.MAX_INFO_TX)
        self.replyData.setUInt8(1)
        self.replyData.setUInt8(self.hdlc.maxInfoTX)
        self.replyData.setUInt8(_HDLCInfo.MAX_INFO_RX)
        self.replyData.setUInt8(1)
        self.replyData.setUInt8(self.hdlc.maxInfoRX)
        self.replyData.setUInt8(_HDLCInfo.WINDOW_SIZE_TX)
        self.replyData.setUInt8(4)
        self.replyData.setUInt32(self.hdlc.windowSizeTX)
        self.replyData.setUInt8(_HDLCInfo.WINDOW_SIZE_RX)
        self.replyData.setUInt8(4)
        self.replyData.setUInt32(self.hdlc.windowSizeRX)
        self.replyData.setUInt8(2, len(self.replyData) - 3)


    def reset(self, connect=False):
        if not connect:
            self.info.clear()
            self.settings.serverAddress = 0
            self.settings.clientAddress = 0
        self.settings.protocolVersion = None
        self.settings.ctoSChallenge = None
        self.settings.stoCChallenge = None
        self.receivedData.clear()
        self.transaction = None
        self.settings.count = 0
        self.settings.index = 0
        self.settings.connected = ConnectionState.NONE
        self.replyData.clear()
        self.settings.authentication = Authentication.NONE
        if self.settings.cipher:
            self.settings.cipher.reset()

    def reset_0(self):
        self.reset(False)

    def handleRequest(self, sr):
        #pylint: disable=too-many-return-statements,broad-except
        """
        Handles client request.

        buff: Received data from the client.
        Returns Response to the request. Response is null if request packet is not complete.
        """
        if not sr.isStreaming() and not sr.data:
            return
        if not self.initialized:
            raise ValueError("Server not Initialized.")
        try:
            if not sr.isStreaming():
                self.receivedData.set(sr.data)
                first = self.settings.serverAddress == 0 and self.settings.clientAddress == 0
                try:
                    GXDLMS.getData(self.settings, self.receivedData, self.info, None)
                except Exception:
                    self.dataReceived = datetime.datetime.now()
                    self.receivedData.size = 0
                    sr.setReply(GXDLMS.getHdlcFrame(self.settings, Command.UNACCEPTABLE_FRAME, self.replyData))
                    return
                if not self.info.complete:
                    return
                self.receivedData.clear()
                if self.info.command == Command.DISCONNECT_REQUEST and (self.settings.connected == ConnectionState.NONE):
                    sr.setReply(GXDLMS.getHdlcFrame(self.settings, Command.DISCONNECT_MODE, self.replyData))
                    self.info.clear()
                    return
                if first or self.info.command == Command.SNRM or (self.settings.interfaceType == InterfaceType.WRAPPER and self.info.command == Command.AARQ):
                    if not self.isTarget(self.settings.serverAddress, self.settings.clientAddress):
                        self.info.clear()
                        return
                if (self.info.moreData & RequestTypes.FRAME) == RequestTypes.FRAME:
                    self.dataReceived = datetime.datetime.now()
                    sr.setReply(GXDLMS.getHdlcFrame(self.settings, self.settings.getReceiverReady(), self.replyData))
                    return
                if self.info.command == Command.NONE:
                    if self.transaction:
                        self.info.command = (self.transaction.command)
                    elif not self.replyData:
                        sr.setReply(GXDLMS.getHdlcFrame(self.settings, self.settings.getReceiverReady(), self.replyData))
                        return
                if self.hdlc and self.hdlc.inactivityTimeout != 0:
                    if self.info.command != Command.SNRM:
                        elapsed = int((datetime.datetime.now() - self.dataReceived)) / 1000
                        if elapsed >= self.hdlc.inactivityTimeout:
                            self.reset()
                            self.dataReceived = 0
                            return
                elif self.wrapper and self.wrapper.inactivityTimeout != 0:
                    if self.info.command != Command.AARQ:
                        elapsed = int((datetime.datetime.now() - self.dataReceived)) / 1000
                        if elapsed >= self.wrapper.inactivityTimeout:
                            self.reset()
                            self.dataReceived = 0
                            return
            else:
                self.info.command = (Command.GENERAL_BLOCK_TRANSFER)
            try:
                sr.setReply(self.handleCommand(self.info.command, self.info.data, sr))
            except Exception:
                self.receivedData.size(0)
                sr.setReply(GXDLMS.getHdlcFrame(self.settings, Command.UNACCEPTABLE_FRAME, self.replyData))
            self.dataReceived = datetime.datetime.now()
            self.info.clear()
        except Exception as e:
            if isinstance(e, (GXDLMSConfirmedServiceError,)):
                sr.setReply(self.reportConfirmedServiceError(e))
                self.transaction = None
                self.settings.setCount(0)
                self.settings.setIndex(0)
                self.info.clear()
                self.receivedData.clear()
            elif self.info.command != Command.NONE:
                sr.setReply(self.reportError(self.info.command, ErrorCode.HARDWARE_FAULT))
                self.transaction = None
                self.settings.setCount(0)
                self.settings.setIndex(0)
                self.info.clear()
                self.receivedData.clear()
            else:
                self.reset()
                if (self.settings.connected & ConnectionState.DLMS) != 0:
                    self.settings.connected = self.settings.connected & ~ConnectionState.DLMS
                    self.onDisconnected(sr.connectionInfo)

    def reportConfirmedServiceError(self, e):
        self.replyData.clear()
        if self.settings.interfaceType == InterfaceType.HDLC or self.settings.interfaceType == InterfaceType.HDLC_WITH_MODE_E:
            GXDLMS.addLLCBytes(self.settings, self.replyData)
        self.replyData.setUInt8(Command.CONFIRMED_SERVICE_ERROR)
        self.replyData.setUInt8(e.confirmedServiceError)
        self.replyData.setUInt8(e.serviceError)
        self.replyData.setUInt8(e.serviceErrorValue)
        if self.settings.interfaceType == InterfaceType.WRAPPER:
            return GXDLMS.getWrapperFrame(self.settings, Command.CONFIRMED_SERVICE_ERROR, self.replyData)
        return GXDLMS.getHdlcFrame(self.settings, int(0), self.replyData)

    def reportError(self, command, error):
        cmd = 0
        if command == Command.READ_REQUEST:
            cmd = Command.READ_RESPONSE
        elif command == Command.WRITE_REQUEST:
            cmd = Command.WRITE_RESPONSE
        elif command == Command.GET_REQUEST:
            cmd = Command.GET_RESPONSE
        elif command == Command.SET_REQUEST:
            cmd = Command.SET_RESPONSE
        elif command == Command.METHOD_REQUEST:
            cmd = Command.METHOD_RESPONSE
        else:
            cmd = Command.NONE
        if self.settings.getUseLogicalNameReferencing():
            p = GXDLMSLNParameters(self.settings, 0, cmd, 1, None, None, error)
            GXDLMS.getLNPdu(p, self.replyData)
        else:
            bb = GXByteBuffer()
            bb.setUInt8(error)
            p2 = GXDLMSSNParameters(self.settings, cmd, 1, 1, None, bb)
            GXDLMS.getSNPdu(p2, self.replyData)
        if self.settings.interfaceType == InterfaceType.WRAPPER:
            return GXDLMS.getWrapperFrame(self.settings, command, self.replyData)
        return GXDLMS.getHdlcFrame(self.settings, int(0), self.replyData)

    def handleCommand(self, cmd, data, sr):
        frame_ = 0
        if self.replyData:
            frame_ = self.settings.getNextSend(False)
        if cmd == Command.ACCESS_REQUEST:
            GXDLMSLNCommandHandler.handleAccessRequest(self.settings, self, data, self.replyData, None)
        elif cmd == Command.SET_REQUEST:
            GXDLMSLNCommandHandler.handleSetRequest(self.settings, self, data, self.replyData, None)
        elif cmd == Command.WRITE_REQUEST:
            GXDLMSSNCommandHandler.handleWriteRequest(self.settings, self, data, self.replyData, None)
        elif cmd == Command.GET_REQUEST:
            if data:
                GXDLMSLNCommandHandler.handleGetRequest(self.settings, self, data, self.replyData, None)
        elif cmd == Command.READ_REQUEST:
            GXDLMSSNCommandHandler.handleReadRequest(self.settings, self, data, self.replyData, None)
        elif cmd == Command.METHOD_REQUEST:
            GXDLMSLNCommandHandler.handleMethodRequest(self.settings, self, data, sr.getConnectionInfo(), self.replyData, None)
        elif cmd == Command.SNRM:
            self.handleSnrmRequest(data)
            frame_ = int(Command.UA)
        elif cmd == Command.AARQ:
            self.handleAarqRequest(data, sr.getConnectionInfo())
            if (self.settings.connected & ConnectionState.DLMS) != 0:
                self.onConnected(sr.getConnectionInfo())
        elif cmd == Command.RELEASE_REQUEST:
            self.handleReleaseRequest(data)
            if (self.settings.connected & ConnectionState.DLMS) != 0:
                self.settings.connected = self.settings.connected & ~ConnectionState.DLMS
                self.onDisconnected(sr.getConnectionInfo())
        elif cmd == Command.DISCONNECT_REQUEST:
            self.generateDisconnectRequest()
            if (self.settings.connected & ConnectionState.DLMS) != 0:
                self.onDisconnected(sr.getConnectionInfo())
            self.settings.connected = ConnectionState.HDLC
            frame_ = Command.UA
        elif cmd == Command.GENERAL_BLOCK_TRANSFER:
            if not self.handleGeneralBlockTransfer(data, sr):
                return None
        elif cmd == Command.NONE:
            pass
        else:
            raise Exception("Invalid command: " + str(cmd))
        if self.settings.interfaceType == InterfaceType.WRAPPER:
            reply = GXDLMS.getWrapperFrame(self.settings, cmd, self.replyData)
        else:
            reply = GXDLMS.getHdlcFrame(self.settings, frame_, self.replyData)
        if cmd == Command.DISCONNECT_REQUEST or (self.settings.interfaceType == InterfaceType.WRAPPER and cmd == Command.RELEASE_REQUEST):
            self.reset()
        return reply

    def handleGeneralBlockTransfer(self, data, sr):
        if self.transaction:
            if self.transaction.command == Command.GET_REQUEST:
                if sr.count == 0:
                    self.settings.setBlockNumberAck(self.settings.blockNumberAck + 1)
                    sr.setCount(self.settings.gbtWindowSize)
                GXDLMSLNCommandHandler.getRequestNextDataBlock(self.settings, 0, self, data, self.replyData, None, True)
                if sr.count != 0:
                    sr.setCount(sr.getCount() - 1)
                if not self.transaction:
                    sr.setCount(0)
            else:
                bc = data.getUInt8()
                blockNumber = data.getUInt16()
                blockNumberAck = data.getUInt16()
                len_ = _GXCommon.getObjectCount(data)
                if len_ > len(data) - data.position:
                    self.replyData.set(self.generateConfirmedServiceError(ConfirmedServiceError.INITIATE_ERROR, ServiceError.SERVICE, Service.UNSUPPORTED))
                else:
                    self.transaction.data.set(data)
                    igonoreAck = (bc & 0x40) != 0 and (blockNumberAck * self.settings.gbtWindowSize) + 1 > blockNumber
                    gbtWindowSize = self.settings.gbtWindowSize
                    bn = self.settings.blockIndex
                    if (bc & 0x80) != 0:
                        self.handleCommand(self.transaction.command, self.transaction.data, sr)
                        self.transaction = None
                        igonoreAck = False
                        gbtWindowSize = 1
                    if igonoreAck:
                        return False
                    self.replyData.setUInt8(Command.GENERAL_BLOCK_TRANSFER)
                    self.replyData.setUInt8(int((0x80 | gbtWindowSize)))
                    self.settings.blockIndex = self.settings.blockIndex + 1
                    self.replyData.setUInt16(bn)
                    self.replyData.setUInt16(blockNumber)
                    self.replyData.setUInt8(0)
        else:
            bc = data.getUInt8()
            blockNumber = data.getUInt16()
            blockNumberAck = data.getUInt16()
            len_ = _GXCommon.getObjectCount(data)
            if len_ > len(data) - data.position:
                self.replyData.set(self.generateConfirmedServiceError(ConfirmedServiceError.INITIATE_ERROR, ServiceError.SERVICE, Service.UNSUPPORTED))
            else:
                self.transaction = GXDLMSLongTransaction(None, data.getUInt8(), data)
                self.replyData.setUInt8(Command.GENERAL_BLOCK_TRANSFER)
                self.replyData.setUInt8((0x80 | self.settings.gbtWindowSize))
                self.replyData.setUInt16(blockNumber)
                blockNumberAck += 1
                self.replyData.setUInt16(blockNumberAck)
                self.replyData.setUInt8(0)
        return True

    @classmethod
    def generateConfirmedServiceError(cls, service, type_, code_):
        return [Command.CONFIRMED_SERVICE_ERROR, service, type_, code_]



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXByteBuffer import GXByteBuffer
from .enums import DataType, RequestTypes, Command
from .GXDLMSException import GXDLMSException

# pylint: disable=bad-option-value,old-style-class,too-few-public-methods,too-many-instance-attributes
class GXReplyData:
    #
    #      Constructor.
    #
    #      @param more
    #                 Is more data available.
    #      @param cmd
    #                 Received command.
    #      @param buff
    #                 Received data.
    #      @param forComplete
    #                 Is frame complete.
    #      @param err
    #                 Received error ID.
    # pylint: disable=too-many-arguments
    def __init__(self, more=RequestTypes.NONE, cmd=Command.NONE, buff=None, forComplete=False, err=0):
        # Is more data available.
        self.moreData = more
        # Received command.
        self.command = cmd
        # Received data.
        self.data = buff
        if not self.data:
            self.data = GXByteBuffer()
        # Is frame complete.
        self.complete = forComplete
        # Received error.
        self.error = err
        self.value = None
        # Is received frame echo.
        self.echo = False
        # Received command type.
        self.commandType = 0
        # HDLC frame ID.
        self.frameId = 0
        # Read value.
        self.dataValue = None
        # Expected count of element in the array.
        self.totalCount = 0
        # Last read position.  This is used in peek to solve how far data
        # is read.
        self.readPosition = 0
        # Packet length.
        self.packetLength = 0
        # Try get value.
        self.peek = False
        # Cipher index is position where data is decrypted.
        self.cipherIndex = 0
        # Data notification date time.
        self.time = None
        # XML settings.
        self.xml = None
        # Invoke ID.
        self.invokeId = 0
        # GBT block number.
        self.blockNumber = 0
        # GBT block number ACK.
        self.blockNumberAck = 0
        # Is GBT streaming in use.
        self.streaming = False
        # GBT Window size.  This is for internal use.
        self.gbtWindowSize = 0
        # Client address of the notification message.  Notification
        # message sets
        # this.
        self.targetAddress = 0
        # Server address of the notification message.  Notification
        # message sets
        # this.
        self.sourceAddress = 0
        # Gateway information.
        self.gateway = None
        # Data type.
        self.valueType = DataType.NONE

    def clear(self):
        """"
        Reset data values to default.
        """
        self.moreData = RequestTypes.NONE
        self.command = Command.NONE
        self.commandType = 0
        self.data.capacity = 0
        self.complete = False
        self.error = 0
        self.totalCount = 0
        self.dataValue = None
        self.readPosition = 0
        self.packetLength = 0
        self.valueType = DataType.NONE
        self.cipherIndex = 0
        self.cipheredCommand = 0
        self.time = None
        if self.xml:
            self.xml.xml = ""
        self.invokeId = 0
        self.value = None

    def isMoreData(self):
        """
        Is more data available.
        """
        return self.moreData != RequestTypes.NONE and self.error == 0

    #
    # Is notify message.
    #
    def isNotify(self):
        return self.command == Command.EVENT_NOTIFICATION or self.command == Command.DATA_NOTIFICATION or self.command == Command.INFORMATION_REPORT

    #
    #      Is frame complete.
    #
    # Returns true if frame is complete or false if bytes is
    #      missing.
    #
    def isComplete(self):
        return self.complete

    #
    #      Get Received error.  Value is zero if no error has occurred.
    #
    # Received error.
    #
    def getError(self):
        return self.error

    def getErrorMessage(self):
        return GXDLMSException.getDescription(self.error)

    #
    #      Get total count of element in the array.  If this method is used
    #      peek must
    #      be set true.
    #
    # Count of element in the array.
    #      @see #setPeek
    #      @see #getCount
    #
    def getTotalCount(self):
        return self.totalCount

    #
    #      Get count of read elements.  If this method is used peek must be set
    #      true.
    #
    # Count of read elements.
    #      @see #setPeek
    #      @see #getTotalCount
    #
    def getCount(self):
        if isinstance(self.dataValue, list):
            return len(self.dataValue)
        return 0

    #
    # Is GBT streaming.
    #
    def isStreaming(self):
        return (self.moreData & RequestTypes.FRAME) == 0 and self.streaming and (self.blockNumberAck * self.gbtWindowSize) + 1 > self.blockNumber

    def __str__(self):
        if self.xml:
            return self.xml
        if self.data is None:
            return ""
        return str(self.data)



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from .enums import PppSetupLcpOptionType
from ..internal._GXCommon import _GXCommon
from ..GXByteBuffer import GXByteBuffer
from ..enums import ErrorCode, ObjectType, DataType
from .GXDLMSPppSetupLcpOption import GXDLMSPppSetupLcpOption
from .GXDLMSPppSetupIPCPOption import GXDLMSPppSetupIPCPOption
from .enums import PppSetupIPCPOptionType

# pylint: disable=too-many-instance-attributes
class GXDLMSPppSetup(GXDLMSObject, IGXDLMSBase):
    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSPppSetup
    """

    #
    # Constructor.
    #
    # @param ln
    # Logical Name of the object.
    # @param sn
    # Short Name of the object.
    #
    def __init__(self, ln=None, sn=0):
        super(GXDLMSPppSetup, self).__init__(ObjectType.PPP_SETUP, ln, sn)
        self.ipcpOptions = list()
        self.phyReference = None
        self.lcpOptions = list()
        # PPP authentication procedure user name.
        self.userName = None
        # PPP authentication procedure password.
        self.password = None
        self.authentication = None

    def getValues(self):
        return [self.logicalName,
                self.phyReference,
                self.lcpOptions,
                self.ipcpOptions,
                [self.userName, self.password]]

    #
    # Returns collection of attributes to read.  If attribute is static
    #      and
    # already read or device is returned HW error it is not returned.
    #
    def getAttributeIndexToRead(self, all_):
        attributes = list()
        #  LN is static and read only once.
        if all_ or not self.logicalName:
            attributes.append(1)
        #  PHYReference
        if all_ or not self.isRead(2):
            attributes.append(2)
        #  LCPOptions
        if all_ or not self.isRead(3):
            attributes.append(3)
        #  IPCPOptions
        if all_ or not self.isRead(4):
            attributes.append(4)
        #  PPPAuthentication
        if all_ or not self.isRead(5):
            attributes.append(5)
        return attributes

    #
    # Returns amount of attributes.
    #
    def getAttributeCount(self):
        return 5

    #
    # Returns amount of methods.
    #
    def getMethodCount(self):
        return 0

    def getDataType(self, index):
        if index == 1:
            return DataType.OCTET_STRING
        if index == 2:
            return DataType.OCTET_STRING
        if index == 3:
            return DataType.ARRAY
        if index == 4:
            return DataType.ARRAY
        if index == 5:
            if not self.userName:
                return DataType.NONE
            return DataType.STRUCTURE
        raise ValueError("getDataType failed. Invalid attribute index.")

    #
    # Returns value of given attribute.
    #
    def getValue(self, settings, e):
        if e.index == 1:
            ret = _GXCommon.logicalNameToBytes(self.logicalName)
        elif e.index == 2:
            ret = _GXCommon.logicalNameToBytes(self.phyReference)
        elif e.index == 3:
            data = GXByteBuffer()
            data.setUInt8(DataType.ARRAY)
            if not self.lcpOptions:
                data.setUInt8(0)
            else:
                data.setUInt8(len(self.lcpOptions))
                for it in self.lcpOptions:
                    data.setUInt8(DataType.STRUCTURE)
                    data.setUInt8(3)
                    _GXCommon.setData(settings, data, DataType.UINT8, it.type_.value)
                    _GXCommon.setData(settings, data, DataType.UINT8, it.length)
                    _GXCommon.setData(settings, data, _GXCommon.getDLMSDataType(it.data), it.data)
            ret = data.array()
        elif e.index == 4:
            data = GXByteBuffer()
            data.setUInt8(DataType.ARRAY)
            if not self.ipcpOptions:
                data.setUInt8(0)
            else:
                data.setUInt8(len(self.ipcpOptions))
                for it in self.ipcpOptions:
                    data.setUInt8(DataType.STRUCTURE)
                    data.setUInt8(3)
                    _GXCommon.setData(settings, data, DataType.UINT8, it.type_.value)
                    _GXCommon.setData(settings, data, DataType.UINT8, it.length)
                    _GXCommon.setData(settings, data, _GXCommon.getDLMSDataType(it.data), it.data)
            ret = data.array()
        elif e.index == 5:
            if self.userName:
                data = GXByteBuffer()
                data.setUInt8(DataType.STRUCTURE)
                data.setUInt8(2)
                _GXCommon.setData(settings, data, DataType.OCTET_STRING, self.userName)
                _GXCommon.setData(settings, data, DataType.OCTET_STRING, self.password)
                ret = data.array()
            else:
                ret = None
        else:
            e.error = ErrorCode.READ_WRITE_DENIED
        return ret

    #
    # Set value of given attribute.
    #
    def setValue(self, settings, e):
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            self.phyReference = _GXCommon.toLogicalName(e.value)
        elif e.index == 3:
            self.lcpOptions = []
            if e.value:
                for item in e.value:
                    it1 = GXDLMSPppSetupLcpOption()
                    it1.type_ = PppSetupLcpOptionType(item[0])
                    it1.length = item[1]
                    it1.data = item[2]
                    self.lcpOptions.append(it1)
        elif e.index == 4:
            self.ipcpOptions = []
            if e.value:
                for item in e.value:
                    it2 = GXDLMSPppSetupIPCPOption()
                    it2.type_ = item[0]
                    it2.length = item[1]
                    it2.data = item[2]
                    self.ipcpOptions.append(it2)
        elif e.index == 5:
            if e.value:
                self.userName = e.value[0]
                self.password = e.value[1]
            else:
                self.userName = None
                self.password = None
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    def load(self, reader):
        self.phyReference = reader.readElementContentAsString("PHYReference")
        self.lcpOptions = []
        if reader.isStartElement("LCPOptions", True):
            while reader.isStartElement("Item", True):
                it1 = GXDLMSPppSetupLcpOption()
                it1.type_ = PppSetupLcpOptionType(reader.readElementContentAsInt("Type"))
                it1.length = reader.readElementContentAsInt("Length")
                it1.data = reader.readElementContentAsObject("Data", None)
                self.lcpOptions.append(it1)
            reader.readEndElement("LCPOptions")

        self.ipcpOptions = []
        if reader.isStartElement("IPCPOptions", True):
            while reader.isStartElement("Item", True):
                it2 = GXDLMSPppSetupIPCPOption()
                it2.type_ = PppSetupIPCPOptionType(reader.readElementContentAsInt("Type"))
                it2.length = reader.readElementContentAsInt("Length")
                it2.data = reader.readElementContentAsObject("Data", None)
                self.ipcpOptions.append(it2)
            reader.readEndElement("IPCPOptions")
        self.userName = GXByteBuffer.hexToBytes(reader.readElementContentAsString("UserName"))
        self.password = GXByteBuffer.hexToBytes(reader.readElementContentAsString("Password"))

    def save(self, writer):
        writer.writeElementString("PHYReference", self.phyReference)
        writer.writeStartElement("LCPOptions")
        if self.lcpOptions:
            for it in self.lcpOptions:
                writer.writeStartElement("Item")
                writer.writeElementString("Type", it.type_)
                writer.writeElementString("Length", it.length)
                writer.writeElementObject("Data", it.data)
                writer.writeEndElement()
        writer.writeEndElement()
        writer.writeStartElement("IPCPOptions")
        if self.ipcpOptions:
            for it in self.ipcpOptions:
                writer.writeStartElement("Item")
                writer.writeElementString("Type", it.type_)
                writer.writeElementString("Length", it.length)
                writer.writeElementObject("Data", it.data)
                writer.writeEndElement()
        writer.writeEndElement()
        writer.writeElementString("UserName", GXByteBuffer.hex(self.userName))
        writer.writeElementString("Password", GXByteBuffer.hex(self.password))



from .ActionRequestType import ActionRequestType
from .enums import Command, DataType, Initiate, LoadDataSet, Task, Access, Definition, ApplicationReference, HardwareResource, VdeStateError, StateError, ExceptionServiceError
from .TranslatorGeneralTags import TranslatorGeneralTags
from .TranslatorTags import TranslatorTags
from .SingleReadResponse import SingleReadResponse
from .VariableAccessSpecification import VariableAccessSpecification
from .GetCommandType import GetCommandType
from .SetRequestType import SetRequestType
from .ActionResponseType import ActionResponseType
from .SetResponseType import SetResponseType
from .enums import AccessServiceCommandType, Conformance, ErrorCode
from .internal._GXCommon import _GXCommon
from .ServiceError import ServiceError
from .enums.Service import Service
from .ReleaseResponseReason import ReleaseResponseReason
from .ReleaseRequestReason import ReleaseRequestReason

#pylint: disable=bad-option-value,old-style-class,too-many-public-methods,unused-private-member
class TranslatorStandardTags:
    # Constructor.
    def __init__(self):
        pass

    #
    # Get general tags.
    #
    # @param type
    # @param list
    #
    @classmethod
    def getGeneralTags(cls, list_):
        list_[Command.SNRM] = "Snrm"
        list_[Command.UNACCEPTABLE_FRAME] = "UnacceptableFrame"
        list_[Command.DISCONNECT_MODE] = "DisconnectMode"
        list_[Command.UA] = "Ua"
        list_[Command.AARQ] = "aarq"
        list_[Command.AARE] = "aare"
        list_[TranslatorGeneralTags.APPLICATION_CONTEXT_NAME] = "application-context-name"
        list_[Command.INITIATE_RESPONSE] = "initiateResponse"
        list_[Command.INITIATE_REQUEST] = "initiateRequest"
        list_[TranslatorGeneralTags.NEGOTIATED_QUALITY_OF_SERVICE] = "negotiated-quality-of-service"
        list_[TranslatorGeneralTags.PROPOSED_QUALITY_OF_SERVICE] = "proposed-quality-of-service"
        list_[TranslatorGeneralTags.PROPOSED_DLMS_VERSION_NUMBER] = "proposed-dlms-version-number"
        list_[TranslatorGeneralTags.PROPOSED_MAX_PDU_SIZE] = "client-max-receive-pdu-size"
        list_[TranslatorGeneralTags.PROPOSED_CONFORMANCE] = "proposed-conformance"
        list_[TranslatorGeneralTags.VAA_NAME] = "vaa-name"
        list_[TranslatorGeneralTags.NEGOTIATED_CONFORMANCE] = "negotiated-conformance"
        list_[TranslatorGeneralTags.NEGOTIATED_DLMS_VERSION_NUMBER] = "negotiated-dlms-version-number"
        list_[TranslatorGeneralTags.NEGOTIATED_MAX_PDU_SIZE] = "server-max-receive-pdu-size"
        list_[TranslatorGeneralTags.CONFORMANCE_BIT] = "ConformanceBit"
        list_[TranslatorGeneralTags.SENDER_ACSE_REQUIREMENTS] = "sender-acse-requirements"
        list_[TranslatorGeneralTags.RESPONDER_ACSE_REQUIREMENT] = "responder-acse-requirements"
        list_[TranslatorGeneralTags.RESPONDING_MECHANISM_NAME] = "mechanism-name"
        list_[TranslatorGeneralTags.CALLING_MECHANISM_NAME] = "mechanism-name"
        list_[TranslatorGeneralTags.CALLING_AUTHENTICATION] = "calling-authentication-value"
        list_[TranslatorGeneralTags.RESPONDING_AUTHENTICATION] = "responding-authentication-value"
        list_[Command.RELEASE_REQUEST] = "rlrq"
        list_[Command.RELEASE_RESPONSE] = "rlre"
        list_[Command.DISCONNECT_REQUEST] = "Disc"
        list_[TranslatorGeneralTags.ASSOCIATION_RESULT] = "result"
        list_[TranslatorGeneralTags.RESULT_SOURCE_DIAGNOSTIC] = "result-source-diagnostic"
        list_[TranslatorGeneralTags.ACSE_SERVICE_USER] = "acse-service-user"
        list_[TranslatorGeneralTags.CALLING_AP_TITLE] = "CallingAPTitle"
        list_[TranslatorGeneralTags.RESPONDING_AP_TITLE] = "RespondingAPTitle"
        list_[TranslatorGeneralTags.CHAR_STRING] = "charstring"
        list_[TranslatorGeneralTags.DEDICATED_KEY] = "dedicated-key"
        list_[TranslatorTags.RESPONSE_ALLOWED] = "response-allowed"
        list_[TranslatorGeneralTags.USER_INFORMATION] = "user-information"
        list_[Command.CONFIRMED_SERVICE_ERROR] = "confirmedServiceError"
        list_[Command.INFORMATION_REPORT] = "informationReportRequest"
        list_[Command.EVENT_NOTIFICATION] = "event-notification-request"
        list_[TranslatorGeneralTags.CALLING_AE_INVOCATION_ID] = "calling-AE-invocation-id"
        list_[TranslatorGeneralTags.CALLED_AE_INVOCATION_ID] = "called-AE-invocation-id"
        list_[TranslatorGeneralTags.RESPONDING_AE_INVOCATION_ID] = "responding-AE-invocation-id"
        list_[Command.EXCEPTION_RESPONSE] = "exception-response"
        list_[TranslatorTags.STATE_ERROR] = "state-error"
        list_[TranslatorTags.SERVICE_ERROR] = "service-error"

    #
    # Get SN tags.
    #
    # @param type
    # @param list
    #
    @classmethod
    def getSnTags(cls, list_):
        list_[Command.READ_REQUEST] = "readRequest"
        list_[Command.WRITE_REQUEST] = "writeRequest"
        list_[Command.WRITE_RESPONSE] = "writeResponse"
        list_[Command.WRITE_REQUEST << 8 | SingleReadResponse.DATA] = "Data"
        list_[Command.READ_REQUEST << 8 | VariableAccessSpecification.VARIABLE_NAME] = "variable-name"
        list_[Command.READ_REQUEST << 8 | VariableAccessSpecification.PARAMETERISED_ACCESS] = "parameterized-access"
        list_[Command.READ_REQUEST << 8 | VariableAccessSpecification.BLOCK_NUMBER_ACCESS] = "BlockNumberAccess"
        list_[Command.WRITE_REQUEST << 8 | VariableAccessSpecification.VARIABLE_NAME] = "variable-name"
        list_[Command.READ_RESPONSE] = "readResponse"
        list_[Command.READ_RESPONSE << 8 | SingleReadResponse.DATA_BLOCK_RESULT] = "DataBlockResult"
        list_[Command.READ_RESPONSE << 8 | SingleReadResponse.DATA] = "data"
        list_[Command.WRITE_RESPONSE << 8 | SingleReadResponse.DATA] = "data"
        list_[Command.READ_RESPONSE << 8 | SingleReadResponse.DATA_ACCESS_ERROR] = "data-access-error"

    #
    # Get LN tags.
    #
    # @param type
    # @param list
    #
    @classmethod
    def getLnTags(cls, list_):
        list_[Command.GET_REQUEST] = "get-request"
        list_[Command.GET_REQUEST << 8 | GetCommandType.NORMAL] = "get-request-normal"
        list_[Command.GET_REQUEST << 8 | GetCommandType.NEXT_DATA_BLOCK] = "get-request-next"
        list_[Command.GET_REQUEST << 8 | GetCommandType.WITH_LIST] = "get-request-with-list"
        list_[Command.SET_REQUEST] = "set-request"
        list_[Command.SET_REQUEST << 8 | SetRequestType.NORMAL] = "set-request-normal"
        list_[Command.SET_REQUEST << 8 | SetRequestType.FIRST_DATA_BLOCK] = "set-request-first-data-block"
        list_[Command.SET_REQUEST << 8 | SetRequestType.WITH_DATA_BLOCK] = "set-request-with-datablock"
        list_[Command.SET_REQUEST << 8 | SetRequestType.WITH_LIST] = "set-request-with-list"
        list_[Command.METHOD_REQUEST] = "action-request"
        list_[Command.METHOD_REQUEST << 8 | ActionRequestType.NORMAL] = "action-request-normal"
        list_[Command.METHOD_REQUEST << 8 | ActionRequestType.NEXT_BLOCK] = "ActionRequestForNextDataBlock"
        list_[Command.METHOD_REQUEST << 8 | ActionRequestType.WITH_LIST] = "action-request-with-list"
        list_[Command.METHOD_RESPONSE] = "action-response"
        list_[Command.METHOD_RESPONSE << 8 | ActionResponseType.NORMAL] = "action-response-normal"
        list_[Command.METHOD_RESPONSE << 8 | ActionResponseType.WITH_BLOCK] = "action-response-with-pblock"
        list_[Command.METHOD_RESPONSE << 8 | ActionResponseType.WITH_LIST] = "action-response-with-list"
        list_[Command.METHOD_RESPONSE << 8 | ActionResponseType.NEXT_BLOCK] = "action-response-next-pblock"
        list_[TranslatorTags.SINGLE_RESPONSE] = "single-response"
        list_[int(Command.DATA_NOTIFICATION)] = "data-notification"
        list_[Command.GET_RESPONSE] = "get-response"
        list_[Command.GET_RESPONSE << 8 | GetCommandType.NORMAL] = "get-response-normal"
        list_[Command.GET_RESPONSE << 8 | GetCommandType.NEXT_DATA_BLOCK] = "get-response-with-data-block"
        list_[Command.GET_RESPONSE << 8 | GetCommandType.WITH_LIST] = "get-response-with-list"
        list_[Command.SET_RESPONSE] = "set-response"
        list_[Command.SET_RESPONSE << 8 | SetResponseType.NORMAL] = "set-response-normal"
        list_[Command.SET_RESPONSE << 8 | SetResponseType.DATA_BLOCK] = "set-response-data-block"
        list_[Command.SET_RESPONSE << 8 | SetResponseType.LAST_DATA_BLOCK] = "set-response-with-last-data-block"
        list_[Command.SET_RESPONSE << 8 | SetResponseType.WITH_LIST] = "set-response-with-list"
        list_[Command.ACCESS_REQUEST] = "access-request"
        list_[(Command.ACCESS_REQUEST) << 8 | AccessServiceCommandType.GET] = "access-request-get"
        list_[(Command.ACCESS_REQUEST) << 8 | AccessServiceCommandType.SET] = "access-request-set"
        list_[(Command.ACCESS_REQUEST) << 8 | AccessServiceCommandType.ACTION] = "access-request-action"
        list_[Command.ACCESS_RESPONSE] = "access-response"
        list_[(Command.ACCESS_RESPONSE) << 8 | AccessServiceCommandType.GET] = "access-response-get"
        list_[(Command.ACCESS_RESPONSE) << 8 | AccessServiceCommandType.SET] = "access-response-set"
        list_[(Command.ACCESS_RESPONSE) << 8 | AccessServiceCommandType.ACTION] = "access-response-action"
        list_[TranslatorTags.ACCESS_REQUEST_BODY] = "access-request-body"
        list_[TranslatorTags.LIST_OF_ACCESS_REQUEST_SPECIFICATION] = "access-request-specification"
        list_[TranslatorTags.ACCESS_REQUEST_SPECIFICATION] = "Access-Request-Specification"
        list_[TranslatorTags.ACCESS_REQUEST_LIST_OF_DATA] = "access-request-list-of-data"
        list_[TranslatorTags.ACCESS_RESPONSE_BODY] = "access-response-body"
        list_[TranslatorTags.LIST_OF_ACCESS_RESPONSE_SPECIFICATION] = "access-response-specification"
        list_[TranslatorTags.ACCESS_RESPONSE_SPECIFICATION] = "Access-Response-Specification"
        list_[TranslatorTags.ACCESS_RESPONSE_LIST_OF_DATA] = "access-response-list-of-data"
        list_[TranslatorTags.SERVICE] = "service"
        list_[TranslatorTags.SERVICE_ERROR] = "service-error"
        list_[Command.GENERAL_BLOCK_TRANSFER] = "general-block-transfer"
        list_[Command.GATEWAY_REQUEST] = "gateway-request"
        list_[Command.GATEWAY_RESPONSE] = "gateway-response"

    #
    # Get glo tags.
    #
    # @param type
    # @param list
    #
    @classmethod
    def getGloTags(cls, list_):
        list_[Command.GLO_INITIATE_REQUEST] = "glo-initiate-request"
        list_[Command.GLO_INITIATE_RESPONSE] = "glo-initiate-response"
        list_[Command.GLO_GET_REQUEST] = "glo-get-request"
        list_[Command.GLO_GET_RESPONSE] = "glo-get-response"
        list_[Command.GLO_SET_REQUEST] = "glo-set-request"
        list_[Command.GLO_SET_RESPONSE] = "glo-set-response"
        list_[Command.GLO_METHOD_REQUEST] = "glo-action-request"
        list_[Command.GLO_METHOD_RESPONSE] = "glo-action-response"
        list_[Command.GLO_READ_REQUEST] = "glo-read-request"
        list_[Command.GLO_READ_RESPONSE] = "glo-read-response"
        list_[Command.GLO_WRITE_REQUEST] = "glo-write-request"
        list_[Command.GLO_WRITE_RESPONSE] = "glo-write-response"
        list_[Command.GENERAL_GLO_CIPHERING] = "general-glo-ciphering"
        list_[Command.GENERAL_CIPHERING] = "general-ciphering"
        list_[Command.GLO_CONFIRMED_SERVICE_ERROR] = "glo-confirmed-service-error"

    #
    # Get ded tags.
    #
    # @param type
    # @param list
    #
    @classmethod
    def getDedTags(cls, list_):
        list_[Command.DED_GET_REQUEST] = "ded-get-request"
        list_[Command.DED_GET_RESPONSE] = "ded-get-response"
        list_[Command.DED_SET_REQUEST] = "ded-set-request"
        list_[Command.DED_SET_RESPONSE] = "ded-set-response"
        list_[Command.DED_METHOD_REQUEST] = "ded-action-request"
        list_[Command.DED_METHOD_RESPONSE] = "ded-action-response"
        list_[Command.GENERAL_DED_CIPHERING] = "general-ded-ciphering"
        list_[Command.DED_CONFIRMED_SERVICE_ERROR] = "ded-confirmed-service-error"

    #
    # Get translator tags.
    #
    # @param type
    # @param list
    #
    @classmethod
    def getTranslatorTags(cls, list_):
        list_[TranslatorTags.WRAPPER] = "Wrapper"
        list_[TranslatorTags.HDLC] = "Hdlc"
        list_[TranslatorTags.PDU_DLMS] = "xDLMS-APDU"
        list_[TranslatorTags.PDU_CSE] = "aCSE-APDU"
        list_[TranslatorTags.TARGET_ADDRESS] = "TargetAddress"
        list_[TranslatorTags.SOURCE_ADDRESS] = "SourceAddress"
        list_[TranslatorTags.FRAME_TYPE] = "FrameType"
        list_[TranslatorTags.LIST_OF_VARIABLE_ACCESS_SPECIFICATION] = "variable-access-specification"
        list_[TranslatorTags.LIST_OF_DATA] = "list-of-data"
        list_[TranslatorTags.SUCCESS] = "Success"
        list_[TranslatorTags.DATA_ACCESS_ERROR] = "data-access-result"
        list_[TranslatorTags.ATTRIBUTE_DESCRIPTOR] = "cosem-attribute-descriptor"
        list_[TranslatorTags.CLASS_ID] = "class-id"
        list_[TranslatorTags.INSTANCE_ID] = "instance-id"
        list_[TranslatorTags.ATTRIBUTE_ID] = "attribute-id"
        list_[TranslatorTags.METHOD_INVOCATION_PARAMETERS] = "method-invocation-parameters"
        list_[TranslatorTags.SELECTOR] = "selector"
        list_[TranslatorTags.PARAMETER] = "parameter"
        list_[TranslatorTags.LAST_BLOCK] = "last-block"
        list_[TranslatorTags.BLOCK_NUMBER] = "block-number"
        list_[TranslatorTags.RAW_DATA] = "raw-data"
        list_[TranslatorTags.METHOD_DESCRIPTOR] = "cosem-method-descriptor"
        list_[TranslatorTags.METHOD_ID] = "method-id"
        list_[TranslatorTags.RESULT] = "result"
        list_[TranslatorTags.RETURN_PARAMETERS] = "return-parameters"
        list_[TranslatorTags.ACCESS_SELECTION] = "access-selection"
        list_[TranslatorTags.VALUE] = "value"
        list_[TranslatorTags.ACCESS_SELECTOR] = "access-selector"
        list_[TranslatorTags.ACCESS_PARAMETERS] = "access-parameters"
        list_[TranslatorTags.ATTRIBUTE_DESCRIPTOR_LIST] = "attribute-descriptor-list"
        list_[TranslatorTags.ATTRIBUTE_DESCRIPTOR_WITH_SELECTION] = "Cosem-Attribute-Descriptor-With-Selection"
        list_[TranslatorTags.READ_DATA_BLOCK_ACCESS] = "ReadDataBlockAccess"
        list_[TranslatorTags.WRITE_DATA_BLOCK_ACCESS] = "WriteDataBlockAccess"
        list_[TranslatorTags.DATA] = "data"
        list_[TranslatorTags.INVOKE_ID] = "invoke-id-and-priority"
        list_[TranslatorTags.LONG_INVOKE_ID] = "long-invoke-id-and-priority"
        list_[TranslatorTags.DATE_TIME] = "date-time"
        list_[TranslatorTags.CURRENT_TIME] = "current-time"
        list_[TranslatorTags.TIME] = "time"
        list_[TranslatorTags.REASON] = "reason"
        list_[TranslatorTags.VARIABLE_ACCESS_SPECIFICATION] = "Variable-Access-Specification"
        list_[TranslatorTags.CHOICE] = "CHOICE"
        list_[TranslatorTags.NOTIFICATION_BODY] = "notification-body"
        list_[TranslatorTags.DATA_VALUE] = "data-value"
        list_[TranslatorTags.INITIATE_ERROR] = "initiateError"
        list_[TranslatorTags.CIPHERED_SERVICE] = "ciphered-content"
        list_[TranslatorTags.SYSTEM_TITLE] = "system-title"
        list_[TranslatorTags.DATA_BLOCK] = "datablock"
        list_[TranslatorTags.TRANSACTION_ID] = "transaction-id"
        list_[TranslatorTags.ORIGINATOR_SYSTEM_TITLE] = "originator-system-title"
        list_[TranslatorTags.RECIPIENT_SYSTEM_TITLE] = "recipient-system-title"
        list_[TranslatorTags.OTHER_INFORMATION] = "other-information"
        list_[TranslatorTags.KEY_INFO] = "key-info"
        list_[TranslatorTags.CIPHERED_CONTENT] = "ciphered-content"
        list_[TranslatorTags.AGREED_KEY] = "agreed-key"
        list_[TranslatorTags.KEY_PARAMETERS] = "key-parameters"
        list_[TranslatorTags.KEY_CIPHERED_DATA] = "key-ciphered-data"
        list_[TranslatorTags.ATTRIBUTE_VALUE] = "attribute-value"
        list_[TranslatorTags.MAX_INFO_RX] = "MaxInfoRX"
        list_[TranslatorTags.MAX_INFO_TX] = "MaxInfoTX"
        list_[TranslatorTags.WINDOW_SIZE_RX] = "WindowSizeRX"
        list_[TranslatorTags.WINDOW_SIZE_TX] = "WindowSizeTX"
        list_[TranslatorTags.VALUE_LIST] = "value-list"
        list_[TranslatorTags.DATA_ACCESS_RESULT] = "data-access-result"
        list_[TranslatorTags.BLOCK_CONTROL] = "block-control"
        list_[TranslatorTags.BLOCK_NUMBER_ACK] = "block-number-ack"
        list_[TranslatorTags.BLOCK_DATA] = "block-data"
        list_[TranslatorTags.CONTENTS_DESCRIPTION] = "contents-description"
        list_[TranslatorTags.ARRAY_CONTENTS] = "array-contents"
        list_[TranslatorTags.NETWORK_ID] = "network-id"
        list_[TranslatorTags.PHYSICAL_DEVICE_ADDRESS] = "physical-device-address"
        list_[TranslatorTags.PROTOCOL_VERSION] = "protocol-version"
        list_[TranslatorTags.CALLED_AP_TITLE] = "called-ap-title"
        list_[TranslatorTags.CALLED_AP_INVOCATION_ID] = "called-ap-invocation-id"
        list_[TranslatorTags.CALLED_AE_INVOCATION_ID] = "called-ae-invocation-id"
        list_[TranslatorTags.CALLING_AP_INVOCATION_ID] = "calling-ap-invocation-id"
        list_[TranslatorTags.CALLED_AE_QUALIFIER] = "called-ae-qualifier"

    @classmethod
    def getDataTypeTags(cls, list_):
        list_[_GXCommon.DATA_TYPE_OFFSET + DataType.NONE] = "null-data"
        list_[_GXCommon.DATA_TYPE_OFFSET + DataType.ARRAY] = "array"
        list_[_GXCommon.DATA_TYPE_OFFSET + DataType.BCD] = "bcd"
        list_[_GXCommon.DATA_TYPE_OFFSET + DataType.BITSTRING] = "bit-string"
        list_[_GXCommon.DATA_TYPE_OFFSET + DataType.BOOLEAN] = "boolean"
        list_[_GXCommon.DATA_TYPE_OFFSET + DataType.COMPACT_ARRAY] = "compact-array"
        list_[_GXCommon.DATA_TYPE_OFFSET + DataType.DATE] = "date"
        list_[_GXCommon.DATA_TYPE_OFFSET + DataType.DATETIME] = "date-time"
        list_[_GXCommon.DATA_TYPE_OFFSET + DataType.ENUM] = "enum"
        list_[_GXCommon.DATA_TYPE_OFFSET + DataType.FLOAT32] = "float32"
        list_[_GXCommon.DATA_TYPE_OFFSET + DataType.FLOAT64] = "float64,"
        list_[_GXCommon.DATA_TYPE_OFFSET + DataType.INT16] = "long"
        list_[_GXCommon.DATA_TYPE_OFFSET + DataType.INT32] = "double-long"
        list_[_GXCommon.DATA_TYPE_OFFSET + DataType.INT64] = "long64"
        list_[_GXCommon.DATA_TYPE_OFFSET + DataType.INT8] = "integer"
        list_[_GXCommon.DATA_TYPE_OFFSET + DataType.OCTET_STRING] = "octet-string"
        list_[_GXCommon.DATA_TYPE_OFFSET + DataType.STRING] = "visible-string"
        list_[_GXCommon.DATA_TYPE_OFFSET + DataType.STRING_UTF8] = "utf8-string"
        list_[_GXCommon.DATA_TYPE_OFFSET + DataType.STRUCTURE] = "structure"
        list_[_GXCommon.DATA_TYPE_OFFSET + DataType.TIME] = "time"
        list_[_GXCommon.DATA_TYPE_OFFSET + DataType.UINT16] = "long-unsigned"
        list_[_GXCommon.DATA_TYPE_OFFSET + DataType.UINT32] = "double-long-unsigned"
        list_[_GXCommon.DATA_TYPE_OFFSET + DataType.UINT64] = "long64-unsigned"
        list_[_GXCommon.DATA_TYPE_OFFSET + DataType.UINT8] = "unsigned"

    @classmethod
    def errorCodeToString(cls, value):
        value = ErrorCode(value)
        str_ = None
        if value == ErrorCode.ACCESS_VIOLATED:
            str_ = "scope-of-access-violated"
        elif value == ErrorCode.DATA_BLOCK_NUMBER_INVALID:
            str_ = "data-block-number-invalid"
        elif value == ErrorCode.DATA_BLOCK_UNAVAILABLE:
            str_ = "data-block-unavailable"
        elif value == ErrorCode.HARDWARE_FAULT:
            str_ = "hardware-fault"
        elif value == ErrorCode.INCONSISTENT_CLASS:
            str_ = "object-class-inconsistent"
        elif value == ErrorCode.LONG_GET_OR_READ_ABORTED:
            str_ = "long-get-aborted"
        elif value == ErrorCode.LONG_SET_OR_WRITE_ABORTED:
            str_ = "long-set-aborted"
        elif value == ErrorCode.NO_LONG_GET_OR_READ_IN_PROGRESS:
            str_ = "no-long-get-in-progress"
        elif value == ErrorCode.NO_LONG_SET_OR_WRITE_IN_PROGRESS:
            str_ = "no-long-set-in-progress"
        elif value == ErrorCode.OK:
            str_ = "success"
        elif value == ErrorCode.OTHER_REASON:
            str_ = "other-reason"
        elif value == ErrorCode.READ_WRITE_DENIED:
            str_ = "read-write-denied"
        elif value == ErrorCode.TEMPORARY_FAILURE:
            str_ = "temporary-failure"
        elif value == ErrorCode.UNAVAILABLE_OBJECT:
            str_ = "object-unavailable"
        elif value == ErrorCode.UNDEFINED_OBJECT:
            str_ = "object-undefined"
        elif value == ErrorCode.UNMATCHED_TYPE:
            str_ = "type-unmatched"
        else:
            raise ValueError("Error code: " + str(value))
        return str_

    @classmethod
    def value_ofErrorCode(cls, value):
        v = None
        if "scope-of-access-violated".lower() == value.lower():
            v = ErrorCode.ACCESS_VIOLATED
        elif "data-block-number-invalid".lower() == value.lower():
            v = ErrorCode.DATA_BLOCK_NUMBER_INVALID
        elif "data-block-unavailable".lower() == value.lower():
            v = ErrorCode.DATA_BLOCK_UNAVAILABLE
        elif "hardware-fault".lower() == value.lower():
            v = ErrorCode.HARDWARE_FAULT
        elif "object-class-inconsistent".lower() == value.lower():
            v = ErrorCode.INCONSISTENT_CLASS
        elif "long-get-aborted".lower() == value.lower():
            v = ErrorCode.LONG_GET_OR_READ_ABORTED
        elif "long-set-aborted".lower() == value.lower():
            v = ErrorCode.LONG_SET_OR_WRITE_ABORTED
        elif "no-long-get-in-progress".lower() == value.lower():
            v = ErrorCode.NO_LONG_GET_OR_READ_IN_PROGRESS
        elif "no-long-set-in-progress".lower() == value.lower():
            v = ErrorCode.NO_LONG_SET_OR_WRITE_IN_PROGRESS
        elif "success".lower() == value.lower():
            v = ErrorCode.OK
        elif "other-reason".lower() == value.lower():
            v = ErrorCode.OTHER_REASON
        elif "read-write-denied".lower() == value.lower():
            v = ErrorCode.READ_WRITE_DENIED
        elif "temporary-failure".lower() == value.lower():
            v = ErrorCode.TEMPORARY_FAILURE
        elif "object-unavailable".lower() == value.lower():
            v = ErrorCode.UNAVAILABLE_OBJECT
        elif "object-undefined".lower() == value.lower():
            v = ErrorCode.UNDEFINED_OBJECT
        elif "type-unmatched".lower() == value.lower():
            v = ErrorCode.UNMATCHED_TYPE
        else:
            raise ValueError("Error code: " + value)
        return v

    @classmethod
    def __getServiceErrors(cls):
        list_ = dict()
        list_[ServiceError.APPLICATION_REFERENCE] = "application-reference"
        list_[ServiceError.HARDWARE_RESOURCE] = "hardware-resource"
        list_[ServiceError.VDE_STATE_ERROR] = "vde-state-error"
        list_[ServiceError.SERVICE] = "service"
        list_[ServiceError.DEFINITION] = "definition"
        list_[ServiceError.ACCESS] = "access"
        list_[ServiceError.INITIATE] = "initiate"
        list_[ServiceError.LOAD_DATASET] = "load-data-set"
        list_[ServiceError.TASK] = "task"
        return list_

    @classmethod

    def __getApplicationReference(cls):
        list_ = dict()
        list_[ApplicationReference.APPLICATION_CONTEXT_UNSUPPORTED] = "application-context-unsupported"
        list_[ApplicationReference.APPLICATION_REFERENCE_INVALID] = "application-reference-invalid"
        list_[ApplicationReference.APPLICATION_UNREACHABLE] = "application-unreachable"
        list_[ApplicationReference.DECIPHERING_ERROR] = "deciphering-error"
        list_[ApplicationReference.OTHER] = "other"
        list_[ApplicationReference.PROVIDER_COMMUNICATION_ERROR] = "provider-communication-error"
        list_[ApplicationReference.TIME_ELAPSED] = "time-elapsed"
        return list_

    @classmethod
    def __getHardwareResource(cls):
        list_ = dict()
        list_[HardwareResource.MASS_STORAGE_UNAVAILABLE] = "mass-storage-unavailable"
        list_[HardwareResource.MEMORY_UNAVAILABLE] = "memory-unavailable"
        list_[HardwareResource.OTHER] = "other"
        list_[HardwareResource.OTHER_RESOURCE_UNAVAILABLE] = "other-resource-unavailable"
        list_[HardwareResource.PROCESSOR_RESOURCE_UNAVAILABLE] = "processor-resource-unavailable"
        return list_

    @classmethod
    def __getVdeStateError(cls):
        list_ = dict()
        list_[VdeStateError.LOADING_DATASET] = "loading-data-set"
        list_[VdeStateError.NO_DLMS_CONTEXT] = "no-dlms-context"
        list_[VdeStateError.OTHER] = "other"
        list_[VdeStateError.STATUS_INOPERABLE] = "status-inoperable"
        list_[VdeStateError.STATUS_NO_CHANGE] = "status-nochange"
        return list_

    @classmethod
    def __getService(cls):
        list_ = dict()
        list_[Service.OTHER] = "other"
        list_[Service.PDU_SIZE] = "pdu-size"
        list_[Service.UNSUPPORTED] = "service-unsupported"
        return list_

    @classmethod
    def __getDefinition(cls):
        list_ = dict()
        list_[Definition.OBJECT_ATTRIBUTE_INCONSISTENT] = "object-attribute-inconsistent"
        list_[Definition.OBJECT_CLASS_INCONSISTENT] = "object-class-inconsistent"
        list_[Definition.OBJECT_UNDEFINED] = "object-undefined"
        list_[Definition.OTHER] = "other"
        return list_

    @classmethod
    def __getAccess(cls):
        list_ = dict()
        list_[Access.HARDWARE_FAULT] = "hardware-fault"
        list_[Access.OBJECT_ACCESS_INVALID] = "object-access-violated"
        list_[Access.OBJECT_UNAVAILABLE] = "object-unavailable"
        list_[Access.OTHER] = "other"
        list_[Access.SCOPE_OF_ACCESS_VIOLATED] = "scope-of-access-violated"
        return list_

    @classmethod
    def __getInitiate(cls):
        list_ = dict()
        list_[Initiate.DLMS_VERSION_TOO_LOW] = "dlms-version-too-low"
        list_[Initiate.INCOMPATIBLE_CONFORMANCE] = "incompatible-conformance"
        list_[Initiate.OTHER] = "other"
        list_[Initiate.PDU_SIZE_TOO_SHORT] = "pdu-size-too-short"
        list_[Initiate.REFUSED_BY_THE_VDE_HANDLER] = "refused-by-the-VDE-Handler"
        return list_

    @classmethod
    def __getLoadDataSet(cls):
        list_ = dict()
        list_[LoadDataSet.DATASET_NOT_READY] = "data-set-not-ready"
        list_[LoadDataSet.DATASET_SIZE_TOO_LARGE] = "dataset-size-too-large"
        list_[LoadDataSet.INTERPRETATION_FAILURE] = "interpretation-failure"
        list_[LoadDataSet.NOT_AWAITED_SEGMENT] = "not-awaited-segment"
        list_[LoadDataSet.NOT_LOADABLE] = "not-loadable"
        list_[LoadDataSet.OTHER] = "other"
        list_[LoadDataSet.PRIMITIVE_OUT_OF_SEQUENCE] = "primitive-out-of-sequence"
        list_[LoadDataSet.STORAGE_FAILURE] = "storage-failure"
        return list_

    @classmethod
    def __getTask(cls):
        list_ = dict()
        list_[Task.NO_REMOTE_CONTROL] = "no-remote-control"
        list_[Task.OTHER] = "other"
        list_[Task.TI_RUNNING] = "ti-running"
        list_[Task.TI_STOPPED] = "ti-stopped"
        list_[Task.TI_UNUSABLE] = "ti-unusable"
        return list_

    @classmethod
    def getServiceErrorValue(cls, error, value):
        ret = ""
        if error == ServiceError.APPLICATION_REFERENCE:
            ret = cls.__getApplicationReference().get(ApplicationReference(value))
        elif error == ServiceError.HARDWARE_RESOURCE:
            ret = cls.__getHardwareResource().get(HardwareResource(value))
        elif error == ServiceError.VDE_STATE_ERROR:
            ret = cls.__getVdeStateError().get(VdeStateError(value))
        elif error == ServiceError.SERVICE:
            ret = cls.__getService().get(Service(value))
        elif error == ServiceError.DEFINITION:
            ret = cls.__getDefinition().get(Definition(value))
        elif error == ServiceError.ACCESS:
            ret = cls.__getAccess().get(Access(value))
        elif error == ServiceError.INITIATE:
            ret = cls.__getInitiate().get(Initiate(value))
        elif error == ServiceError.LOAD_DATASET:
            ret = cls.__getLoadDataSet().get(LoadDataSet(value))
        elif error == ServiceError.TASK:
            ret = cls.__getTask().get(Task(value))
        elif error == ServiceError.OTHER_ERROR:
            ret = str(value)
        return ret

    #
    # @param error
    #            Service error enumeration value.
    # Service error standard XML tag.
    #
    @classmethod
    def serviceErrorToString(cls, error):
        return cls.__getServiceErrors().get(error)

    #
    # @param value
    #            Service error standard XML tag.
    # Service error enumeration value.
    #
    @classmethod
    def getServiceError(cls, value):
        error = None
        for k, v in cls.__getServiceErrors():
            if value.compareTo(v) == 0:
                error = k
                break
        if error is None:
            raise ValueError()
        return error

    @classmethod
    def __getApplicationReferenceByValue(cls, value):
        ret = None
        for k, v in cls.__getApplicationReference():
            if value.compareTo(v) == 0:
                ret = k
                break
        if ret is None:
            raise ValueError()
        return ret

    @classmethod
    def __getHardwareResourceByValue(cls, value):
        ret = None
        for k, v in cls.__getHardwareResource():
            if value.compareTo(v) == 0:
                ret = k
                break
        if ret is None:
            raise ValueError()
        return ret

    @classmethod
    def __getVdeStateErrorByValue(cls, value):
        ret = None
        for k, v in cls.__getVdeStateError():
            if value.compareTo(v) == 0:
                ret = k
                break
        if ret is None:
            raise ValueError()
        return ret

    @classmethod
    def __getServiceByValue(cls, value):
        ret = None
        for k, v in cls.__getService():
            if value.compareTo(v) == 0:
                ret = k
                break
        if ret is None:
            raise ValueError()
        return ret

    @classmethod
    def __getDefinitionByValue(cls, value):
        ret = None
        for k, v in cls.__getDefinition():
            if value.compareTo(v) == 0:
                ret = k
                break
        if ret is None:
            raise ValueError()
        return ret

    @classmethod
    def __getAccessByValue(cls, value):
        ret = None
        for k, v in cls.__getAccess():
            if value.compareTo(v) == 0:
                ret = k
                break
        if ret is None:
            raise ValueError()
        return ret

    @classmethod
    def __getInitiateByValue(cls, value):
        ret = None
        for k, v in cls.__getInitiate():
            if value.compareTo(v) == 0:
                ret = k
                break
        if ret is None:
            raise ValueError()
        return ret

    @classmethod
    def __getLoadDataSetByValue(cls, value):
        ret = None
        for k, v in cls.__getLoadDataSet():
            if value.compareTo(v) == 0:
                ret = k
                break
        if ret is None:
            raise ValueError()
        return ret

    @classmethod
    def __getTaskByValue(cls, value):
        ret = None
        for k, v in cls.__getTask():
            if value.compareTo(v) == 0:
                ret = k
                break
        if ret is None:
            raise ValueError()
        return ret

    @classmethod
    def getError(cls, serviceError, value):
        ret = 0
        if serviceError == ServiceError.APPLICATION_REFERENCE:
            ret = cls.__getApplicationReferenceByValue(value)
        elif serviceError == ServiceError.HARDWARE_RESOURCE:
            ret = cls.__getHardwareResourceByValue(value)
        elif serviceError == ServiceError.VDE_STATE_ERROR:
            ret = cls.__getVdeStateErrorByValue(value)
        elif serviceError == ServiceError.SERVICE:
            ret = cls.__getServiceByValue(value)
        elif serviceError == ServiceError.DEFINITION:
            ret = cls.__getDefinitionByValue(value)
        elif serviceError == ServiceError.ACCESS:
            ret = cls.__getAccessByValue(value)
        elif serviceError == ServiceError.INITIATE:
            ret = cls.__getInitiateByValue(value)
        elif serviceError == ServiceError.LOAD_DATASET:
            ret = cls.__getLoadDataSetByValue(value)
        elif serviceError == ServiceError.TASK:
            ret = cls.__getTaskByValue(value)
        elif serviceError == ServiceError.OTHER_ERROR:
            ret = int(value)
        return int(ret)

    @classmethod
    def conformancetoString(cls, value):
        str_ = None
        if value == Conformance.ACCESS:
            str_ = "access"
        elif value == Conformance.ACTION:
            str_ = "action"
        elif value == Conformance.ATTRIBUTE_0_SUPPORTED_WITH_GET:
            str_ = "attribute0-supported-with-get"
        elif value == Conformance.ATTRIBUTE_0_SUPPORTED_WITH_SET:
            str_ = "attribute0-supported-with-set"
        elif value == Conformance.BLOCK_TRANSFER_WITH_ACTION:
            str_ = "block-transfer-with-action"
        elif value == Conformance.BLOCK_TRANSFER_WITH_GET_OR_READ:
            str_ = "block-transfer-with-get-or-read"
        elif value == Conformance.BLOCK_TRANSFER_WITH_SET_OR_WRITE:
            str_ = "block-transfer-with-set-or-write"
        elif value == Conformance.DATA_NOTIFICATION:
            str_ = "data-notification"
        elif value == Conformance.EVENT_NOTIFICATION:
            str_ = "event-notification"
        elif value == Conformance.GENERAL_BLOCK_TRANSFER:
            str_ = "general-block-transfer"
        elif value == Conformance.GENERAL_PROTECTION:
            str_ = "general-protection"
        elif value == Conformance.GET:
            str_ = "get"
        elif value == Conformance.INFORMATION_REPORT:
            str_ = "information-report"
        elif value == Conformance.MULTIPLE_REFERENCES:
            str_ = "multiple-references"
        elif value == Conformance.PARAMETERIZED_ACCESS:
            str_ = "parameterized-access"
        elif value == Conformance.PRIORITY_MGMT_SUPPORTED:
            str_ = "priority-mgmt-supported"
        elif value == Conformance.READ:
            str_ = "read"
        elif value == Conformance.RESERVED_SEVEN:
            str_ = "reserved-seven"
        elif value == Conformance.DELTA_VALUE_ENCODING:
            str_ = "delta-value-encoding"
        elif value == Conformance.RESERVED_ZERO:
            str_ = "reserved-zero"
        elif value == Conformance.SELECTIVE_ACCESS:
            str_ = "selective-access"
        elif value == Conformance.SET:
            str_ = "set"
        elif value == Conformance.UN_CONFIRMED_WRITE:
            str_ = "unconfirmed-write"
        elif value == Conformance.WRITE:
            str_ = "write"
        else:
            raise ValueError(str(value))
        return str_

    @classmethod
    def value_ofConformance(cls, value):
        ret = None
        if "access".lower() == value.lower():
            ret = Conformance.ACCESS
        elif "action".lower() == value.lower():
            ret = Conformance.ACTION
        elif "attribute0-supported-with-get".lower() == value.lower():
            ret = Conformance.ATTRIBUTE_0_SUPPORTED_WITH_GET
        elif "attribute0-supported-with-set".lower() == value.lower():
            ret = Conformance.ATTRIBUTE_0_SUPPORTED_WITH_SET
        elif "block-transfer-with-action".lower() == value.lower():
            ret = Conformance.BLOCK_TRANSFER_WITH_ACTION
        elif "block-transfer-with-get-or-read".lower() == value.lower():
            ret = Conformance.BLOCK_TRANSFER_WITH_GET_OR_READ
        elif "block-transfer-with-set-or-write".lower() == value.lower():
            ret = Conformance.BLOCK_TRANSFER_WITH_SET_OR_WRITE
        elif "data-notification".lower() == value.lower():
            ret = Conformance.DATA_NOTIFICATION
        elif "event-notification".lower() == value.lower():
            ret = Conformance.EVENT_NOTIFICATION
        elif "general-block-transfer".lower() == value.lower():
            ret = Conformance.GENERAL_BLOCK_TRANSFER
        elif "general-protection".lower() == value.lower():
            ret = Conformance.GENERAL_PROTECTION
        elif "get".lower() == value.lower():
            ret = Conformance.GET
        elif "information-report".lower() == value.lower():
            ret = Conformance.INFORMATION_REPORT
        elif "multiple-references".lower() == value.lower():
            ret = Conformance.MULTIPLE_REFERENCES
        elif "parameterized-access".lower() == value.lower():
            ret = Conformance.PARAMETERIZED_ACCESS
        elif "priority-mgmt-supported".lower() == value.lower():
            ret = Conformance.PRIORITY_MGMT_SUPPORTED
        elif "read".lower() == value.lower():
            ret = Conformance.READ
        elif "reserved-seven".lower() == value.lower():
            ret = Conformance.RESERVED_SEVEN
        elif "delta-value-encoding".lower() == value.lower():
            ret = Conformance.DELTA_VALUE_ENCODING
        elif "reserved-zero".lower() == value.lower():
            ret = Conformance.RESERVED_ZERO
        elif "selective-access".lower() == value.lower():
            ret = Conformance.SELECTIVE_ACCESS
        elif "set".lower() == value.lower():
            ret = Conformance.SET
        elif "unconfirmed-write".lower() == value.lower():
            ret = Conformance.UN_CONFIRMED_WRITE
        elif "write".lower() == value.lower():
            ret = Conformance.WRITE
        else:
            raise ValueError(value)
        return ret

    @classmethod
    def releaseResponseReasonToString(cls, value):
        str_ = None
        if value == ReleaseResponseReason.NORMAL:
            str_ = "normal"
        elif value == ReleaseResponseReason.NOT_FINISHED:
            str_ = "not-finished"
        elif value == ReleaseResponseReason.USER_DEFINED:
            str_ = "user-defined"
        else:
            raise ValueError(str(value))
        return str_

    @classmethod
    def value_ofReleaseResponseReason(cls, value):
        ret = None
        if "normal".lower() == value.lower():
            ret = ReleaseResponseReason.NORMAL
        elif "not-finished".lower() == value.lower():
            ret = ReleaseResponseReason.NOT_FINISHED
        elif "user-defined".lower() == value.lower():
            ret = ReleaseResponseReason.USER_DEFINED
        else:
            raise ValueError(value)
        return ret

    @classmethod
    def releaseRequestReasonToString(cls, value):
        str_ = None
        if value == ReleaseRequestReason.NORMAL:
            str_ = "normal"
        elif value == ReleaseRequestReason.URGENT:
            str_ = "not-finished"
        elif value == ReleaseRequestReason.USER_DEFINED:
            str_ = "user-defined"
        else:
            raise ValueError(str(value))
        return str_

    @classmethod
    def value_ofReleaseRequestReason(cls, value):
        ret = None
        if "normal".lower() == value.lower():
            ret = ReleaseRequestReason.NORMAL
        elif "not-finished".lower() == value.lower():
            ret = ReleaseRequestReason.URGENT
        elif "user-defined".lower() == value.lower():
            ret = ReleaseRequestReason.USER_DEFINED
        else:
            raise ValueError(value)
        return ret

    @classmethod
    def stateErrorToString(cls, value):
        if value == StateError.SERVICE_NOT_ALLOWED:
            ret = "service-not-allowed"
        elif value == StateError.SERVICE_UNKNOWN:
            ret = "service-unknown"
        else:
            raise ValueError(value)
        return ret

    @classmethod
    def exceptionServiceErrorToString(cls, value):
        if value == ExceptionServiceError.OPERATION_NOT_POSSIBLE:
            ret = "operation-not-possible"
        elif value == ExceptionServiceError.SERVICE_NOT_SUPPORTED:
            ret = "service-not-supported"
        elif value == ExceptionServiceError.OTHER_REASON:
            ret = "other-reason"
        else:
            raise ValueError(value)
        return ret

    @classmethod
    def valueofStateError(cls, value):
        if "service-not-allowed".lower() == value.lower():
            ret = StateError.SERVICE_NOT_ALLOWED
        elif "service-unknown".lower() == value.lower():
            ret = StateError.SERVICE_UNKNOWN
        else:
            raise ValueError(value)
        return ret

    @classmethod
    def valueOfExceptionServiceError(cls, value):
        if "operation-not-possible".lower() == value.lower():
            ret = ExceptionServiceError.OPERATION_NOT_POSSIBLE
        elif "service-not-supported".lower() == value.lower():
            ret = ExceptionServiceError.SERVICE_NOT_SUPPORTED
        elif "other-reason".lower() == value.lower():
            ret = ExceptionServiceError.OTHER_REASON
        else:
            raise ValueError(value)
        return ret



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..enums import ErrorCode
from ..internal._GXCommon import _GXCommon
from ..enums import ObjectType, DataType
from .enums import OpticalProtocolMode, BaudRate, LocalPortResponseTime

# pylint: disable=too-many-instance-attributes
class GXDLMSIECLocalPortSetup(GXDLMSObject, IGXDLMSBase):
    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSIECLocalPortSetup
    """

    def __init__(self, ln="0.0.20.0.0.255", sn=0):
        """
        Constructor.

        ln : Logical Name of the object.
        sn : Short Name of the object.
        """
        GXDLMSObject.__init__(self, ObjectType.IEC_LOCAL_PORT_SETUP, ln, sn)
        self.defaultMode = OpticalProtocolMode.DEFAULT
        self.defaultBaudrate = BaudRate.BAUDRATE_300
        self.proposedBaudrate = BaudRate.BAUDRATE_300
        self.responseTime = LocalPortResponseTime.ms20
        self.password1 = None
        self.password2 = None
        self.password5 = None
        self.deviceAddress = None
        self.version = 1

    def getValues(self):
        return [self.logicalName,
                self.defaultMode,
                self.defaultBaudrate,
                self.proposedBaudrate,
                self.responseTime,
                self.deviceAddress,
                self.password1,
                self.password2,
                self.password5]

    #
    # Returns collection of attributes to read.  If attribute is static and
    # already read or device is returned HW error it is not returned.
    def getAttributeIndexToRead(self, all_):
        attributes = list()
        #  LN is static and read only once.
        if all_ or not self.logicalName:
            attributes.append(1)
        #  DefaultMode
        if all_ or not self.isRead(2):
            attributes.append(2)
        #  DefaultBaudrate
        if all_ or not self.isRead(3):
            attributes.append(3)
        #  ProposedBaudrate
        if all_ or not self.isRead(4):
            attributes.append(4)
        #  ResponseTime
        if all_ or not self.isRead(5):
            attributes.append(5)
        #  DeviceAddress
        if all_ or not self.isRead(6):
            attributes.append(6)
        #  Password1
        if all_ or not self.isRead(7):
            attributes.append(7)
        #  Password2
        if all_ or not self.isRead(8):
            attributes.append(8)
        #  Password5
        if all_ or not self.isRead(9):
            attributes.append(9)
        return attributes

    #
    # Returns amount of attributes.
    #
    def getAttributeCount(self):
        return 9

    #
    # Returns amount of methods.
    #
    def getMethodCount(self):
        return 0

    def getDataType(self, index):
        if index == 1:
            ret = DataType.OCTET_STRING
        elif index == 2:
            ret = DataType.ENUM
        elif index == 3:
            ret = DataType.ENUM
        elif index == 4:
            ret = DataType.ENUM
        elif index == 5:
            ret = DataType.ENUM
        elif index == 6:
            ret = DataType.OCTET_STRING
        elif index == 7:
            ret = DataType.OCTET_STRING
        elif index == 8:
            ret = DataType.OCTET_STRING
        elif index == 9:
            ret = DataType.OCTET_STRING
        else:
            raise ValueError("getDataType failed. Invalid attribute index.")
        return ret
    #
    # Returns value of given attribute.
    #
    def getValue(self, settings, e):
        if e.index == 1:
            ret = _GXCommon.logicalNameToBytes(self.logicalName)
        elif e.index == 2:
            ret = self.defaultMode
        elif e.index == 3:
            ret = self.defaultBaudrate
        elif e.index == 4:
            ret = self.proposedBaudrate
        elif e.index == 5:
            ret = self.responseTime
        elif e.index == 6:
            ret = _GXCommon.getBytes(self.deviceAddress)
        elif e.index == 7:
            ret = _GXCommon.getBytes(self.password1)
        elif e.index == 8:
            ret = _GXCommon.getBytes(self.password2)
        elif e.index == 9:
            ret = _GXCommon.getBytes(self.password5)
        else:
            e.error = ErrorCode.READ_WRITE_DENIED
        return ret

    def setValue(self, settings, e):
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            self.defaultMode = e.value
        elif e.index == 3:
            self.defaultBaudrate = e.value
        elif e.index == 4:
            self.proposedBaudrate = e.value
        elif e.index == 5:
            self.responseTime = e.value
        elif e.index == 6:
            self.deviceAddress = _GXCommon.changeType(settings, e.value, DataType.STRING)
        elif e.index == 7:
            self.password1 = _GXCommon.changeType(settings, e.value, DataType.STRING)
        elif e.index == 8:
            self.password2 = _GXCommon.changeType(settings, e.value, DataType.STRING)
        elif e.index == 9:
            self.password5 = _GXCommon.changeType(settings, e.value, DataType.STRING)
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    def load(self, reader):
        self.defaultMode = reader.readElementContentAsInt("DefaultMode")
        self.defaultBaudrate = reader.readElementContentAsInt("DefaultBaudrate")
        self.proposedBaudrate = reader.readElementContentAsInt("ProposedBaudrate")
        self.responseTime = reader.readElementContentAsInt("ResponseTime")
        self.deviceAddress = reader.readElementContentAsString("DeviceAddress")
        self.password1 = reader.readElementContentAsString("Password1")
        self.password2 = reader.readElementContentAsString("Password2")
        self.password5 = reader.readElementContentAsString("Password5")

    def save(self, writer):
        writer.writeElementString("DefaultMode", int(self.defaultMode))
        writer.writeElementString("DefaultBaudrate", int(self.defaultBaudrate))
        writer.writeElementString("ProposedBaudrate", int(self.proposedBaudrate))
        writer.writeElementString("ResponseTime", int(self.responseTime))
        writer.writeElementString("DeviceAddress", self.deviceAddress)
        writer.writeElementString("Password1", self.password1)
        writer.writeElementString("Password2", self.password2)
        writer.writeElementString("Password5", self.password5)

class GXDLMSIECOpticalPortSetup(GXDLMSIECLocalPortSetup):
    """Obsolete. Use GXDLMSIECLocalPortSetup instead."""



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
import sys
import struct

#pylint: disable=import-error, no-name-in-module
if sys.version_info < (3, 0):
    __base = object
else:
    from collections.abc import Sequence
    __base = Sequence

# pylint: disable=too-many-public-methods,useless-object-inheritance
class GXByteBuffer(__base):
    """
    Byte array class is used to save received bytes.
    """

    __HEX_ARRAY = "0123456789ABCDEFGH"
    __NIBBLE = 4
    __LOW_BYTE_PART = 0x0F
    __ARRAY_CAPACITY = 10

    def __init__(self, value=None):
        """
        Constructor.
        value: Buffer or capacity.
        """
        self._data = bytearray()
        self.__size = 0
        self.__position = 0
        if isinstance(value, (bytearray, bytes)):
            self.setCapacity(len(value))
            self.set(value)
        elif isinstance(value, GXByteBuffer):
            self.setCapacity(value.size - value.position)
            self.set(value)
        elif isinstance(value, int):
            self.setCapacity(value)
        elif isinstance(value, str):
            self.setHexString(value)
        else:
            self.setCapacity(0)

    def clear(self):
        """
        Clear buffer but do not release memory.
        """
        self.position = 0
        self.size = 0

    #
    #      Buffer capacity.
    #
    # Buffer capacity.
    #
    def getCapacity(self):
        if not self._data:
            return 0
        return len(self._data)

    #
    #      Allocate new size for the array in bytes.
    #
    #      @param capacity
    #                 Buffer capacity.
    #
    def setCapacity(self, capacity):
        if capacity == 0:
            self._data = bytearray()
            self.__size = 0
            self.__position = 0
        else:
            if not self._data:
                self._data = bytearray(capacity)
            else:
                tmp = self._data
                self._data = bytearray(capacity)
                if self.size < capacity:
                    self._data[0:self.size] = tmp
                else:
                    self._data[0:capacity] = self._data
                    self.__size = capacity

    capacity = property(getCapacity, setCapacity)
    """Buffer capacity."""

    def getPosition(self):
        return self.__position

    def setPosition(self, value):
        if value < 0 or value > len(self):
            raise ValueError("position")
        self.__position = value

    position = property(getPosition, setPosition)
    """Buffer position."""

    def getSize(self):
        return self.__size

    def setsize(self, value):
        if value < 0 or value > self.capacity:
            raise ValueError("size")
        self.__size = value
        if self.__position > self.__size:
            self.__position = self.__size

    size = property(getSize, setsize)
    """Buffer size."""

    def __len__(self):
        """Buffer size."""
        return self.__size

    def __getitem__(self, i):
        return self._data[i]

    def available(self):
        """Amount of non read bytes in the buffer."""
        return self.size - self.position

    #
    # Get buffer data as byte array.
    #
    def array(self):
        return self.subArray(0, self.size)

    #
    #      Returns sub array from byte buffer.
    #
    #      @param index
    #                 Start index.
    #      @param count
    #                 Byte count.
    # Sub array.
    #
    def subArray(self, index, count):
        if count != 0:
            tmp = bytearray(count)
            tmp[0:count] = self._data[index:index + count]
            return tmp
        return bytearray(0)

    #
    #      Move content from source to destination.
    #
    #      @param srcPos
    #                 Source position.
    #      @param destPos
    #                 Destination position.
    #      @param count
    #                 Item count.
    #
    def move(self, srcPos, destPos, count):
        if count < 0:
            raise ValueError("count")
        if count != 0:
            self._data[destPos:destPos + count] = self._data[srcPos:srcPos + count]
            self.size = destPos + count
            if self.position > self.size:
                self.position = self.size

    #
    #      Remove handled bytes.  This can be used in debugging to remove
    #      handled
    #      bytes.
    #
    def trim(self):
        if self.size == self.position:
            self.size = 0
        else:
            self.move(self.position, 0, self.size - self.position)
        self.position = 0

    #
    #      Push the given byte into this buffer at the current position, and
    #      then
    #      increments the position.
    #
    #      @param item
    #                 The byte to be added.
    #
    def setUInt8(self, item, index=None):
        if index is None:
            self.setUInt8(item, self.size)
            self.size += 1
        else:
            if index >= self.capacity:
                self.capacity = index + self.__ARRAY_CAPACITY
            self._data[index] = item

    #
    #      Push the given byte into this buffer at the current position, and
    #      then
    #      increments the position.
    #
    #      @param item
    #                 The byte to be added.
    #
    def setInt8(self, item, index=None):
        self.setUInt8(item & 0xFF, index)

    def setUInt16(self, item, index=None):
        if index is None:
            self.setUInt16(item, self.size)
            self.size += 2
        else:
            if index + 2 >= self.capacity:
                self.capacity = (index + self.__ARRAY_CAPACITY)
            self._data[index] = int(((item >> 8) & 0xFF))
            self._data[index + 1] = int((item & 0xFF))

    def setUInt32(self, item, index=None):
        if index is None:
            self.setUInt32(item, self.size)
            self.size += 4
        else:
            if index + 4 >= self.capacity:
                self.capacity = index + self.__ARRAY_CAPACITY
            self._data[index] = int(((item >> 24) & 0xFF))
            self._data[index + 1] = int(((item >> 16) & 0xFF))
            self._data[index + 2] = int(((item >> 8) & 0xFF))
            self._data[index + 3] = int((item & 0xFF))

    def setUInt64(self, item, index=None):
        if index is None:
            self.setUInt64(item, self.size)
            self.size += 8
        else:
            if index + 8 >= self.capacity:
                self.capacity = (index + self.__ARRAY_CAPACITY)
            self._data[self.size] = int(((item >> 56) & 0xFF))
            self._data[self.size + 1] = int(((item >> 48) & 0xFF))
            self._data[self.size + 2] = int(((item >> 40) & 0xFF))
            self._data[self.size + 3] = int(((item >> 32) & 0xFF))
            self._data[self.size + 4] = int(((item >> 24) & 0xFF))
            self._data[self.size + 5] = int(((item >> 16) & 0xFF))
            self._data[self.size + 6] = int(((item >> 8) & 0xFF))
            self._data[self.size + 7] = int((item & 0xFF))

    def setFloat(self, value, index=None):
        if index is None:
            self.setFloat(value, self.size)
            self.size += 4
        else:
            self.set(struct.pack("f", value), index)

    def setDouble(self, value, index=None):
        if index is None:
            self.setDouble(value, self.size)
            self.size += 8
        else:
            self.set(struct.pack("d", value), index)

    def getUInt8(self, index=None):
        if index is None:
            index = self.position
            value = self._data[index] & 0xFF
            value = value % 2 ** 8
            self.position += 1
            return value
        if index >= self.size:
            raise ValueError("getUInt8")
        value = self._data[index] & 0xFF
        value = value % 2 ** 8
        return value

    def getInt8(self, index=None):
        if index is None:
            index = self.position
            value = self._data[index]
            value = (value + 2 ** 7) % 2 ** 8 - 2 ** 7
            self.position += 1
            return value
        if index >= self.size:
            raise ValueError("getInt8")
        value = self._data[index]
        value = (value + 2 ** 7) % 2 ** 8 - 2 ** 7
        return value

    def getUInt16(self, index=None):
        if index is None:
            index = self.position
            value = ((self._data[index] & 0xFF) << 8) | (self._data[index + 1] & 0xFF)
            value = value % 2 ** 16
            self.position += 2
            return value
        if index + 2 > self.size:
            raise ValueError("getUInt16")
        value = ((self._data[index] & 0xFF) << 8) | (self._data[index + 1] & 0xFF)
        value = value % 2 ** 16
        return value

    def getInt16(self):
        return (self.getUInt16() + 2 ** 15) % 2 ** 16 - 2 ** 15

    def getInt32(self, index=None):
        if index is None:
            index = self.position
            if index + 4 > self.size:
                raise ValueError("getInt32")
            value = (self._data[index] & 0xFF) << 24 | (self._data[index + 1] & 0xFF) << 16 | (self._data[index + 2] & 0xFF) << 8 | (self._data[index + 3] & 0xFF)
            value = (value + 2 ** 31) % 2 ** 32 - 2 ** 31
            self.position += 4
            return value

        if index + 4 > self.size:
            raise ValueError("getInt32")
        value = (self._data[index] & 0xFF) << 24 | (self._data[index + 1] & 0xFF) << 16 | (self._data[index + 2] & 0xFF) << 8 | (self._data[index + 3] & 0xFF)
        value = (value + 2 ** 31) % 2 ** 32 - 2 ** 31
        return value

    def getUInt32(self, index=None):
        if index is None:
            index = self.position
            self.position += 4
        if index + 4 > self.size:
            raise ValueError("getUInt32")
        value = self._data[index] & 0xFF
        value = value << 24
        value |= (self._data[index + 1] & 0xFF) << 16
        value |= (self._data[index + 2] & 0xFF) << 8
        value |= (self._data[index + 3] & 0xFF)
        value = value % 2 ** 32
        return value

    def getFloat(self):
        tmp = bytearray(4)
        self.get(tmp)
        # Swap bytes.
        tmp2 = tmp[0]
        tmp[0] = tmp[3]
        tmp[3] = tmp2
        tmp2 = tmp[1]
        tmp[1] = tmp[2]
        tmp[2] = tmp2
        return struct.unpack("f", tmp)[0]

    def getDouble(self):
        tmp = bytearray(8)
        self.get(tmp)
        # Swap bytes.
        tmp2 = tmp[0]
        tmp[0] = tmp[7]
        tmp[7] = tmp2
        tmp2 = tmp[1]
        tmp[1] = tmp[6]
        tmp[6] = tmp2
        tmp2 = tmp[2]
        tmp[2] = tmp[5]
        tmp[5] = tmp2
        tmp2 = tmp[3]
        tmp[3] = tmp[4]
        tmp[4] = tmp2
        return struct.unpack("d", tmp)[0]

    def getInt64(self, index=None):
        if index is None:
            index = self.position
            self.position += 8
        value = ((self._data[index] & 0xFF)) << 56
        value |= ((self._data[index + 1] & 0xFF)) << 48
        value |= ((self._data[index + 2] & 0xFF)) << 40
        value |= ((self._data[index + 3] & 0xFF)) << 32
        value |= ((self._data[index + 4] & 0xFF)) << 24
        value |= (self._data[index + 5] & 0xFF) << 16
        value |= (self._data[index + 6] & 0xFF) << 8
        value |= (self._data[index + 7] & 0xFF)
        value = (value + 2 ** 63) % 2 ** 64 - 2 ** 63
        return value

    def getUInt64(self, index=None):
        value = self.getInt64(index)
        return value % 2 ** 64

    #
    #      Check is byte buffer ASCII string.
    #
    #      @param value
    #                 Byte array.
    # Is ASCII string.
    #
    @classmethod
    def isAsciiString(cls, value):
        # pylint: disable=too-many-boolean-expressions
        if value:
            for it in value:
                if (it < 32 or it > 127) and it != '\r' and it != '\n' and it != '\t' and it != 0:
                    return False
        return True

    def getString(self, index, count):
        if index is None and count is None:
            tmp = self._data[0:self.size]
            if self.isAsciiString(tmp):
                str_ = tmp.decode("utf-8").rstrip('\x00')
            else:
                str_ = self.hex(tmp)
            self.position += count
            return str_

        if index + count > self.size:
            raise ValueError("getString")
        tmp = self._data[index:index + count]
        if self.isAsciiString(tmp):
            return tmp.decode("utf-8").rstrip('\x00')
        return self.hex(tmp)

    def set(self, value, index=None, count=None):
        # pylint: disable=protected-access
        if isinstance(value, str):
            value = value.encode()
        if value:
            if index is None:
                if isinstance(value, GXByteBuffer):
                    index = value.position
                else:
                    index = 0
            if count is None:
                count = len(value) - index
            if isinstance(value, GXByteBuffer):
                self.set(value._data, index, count)
                value.position = index + count
            elif value and count != 0:
                if self.size + count > self.capacity:
                    self.capacity = self.size + count + self.__ARRAY_CAPACITY
                self._data[self.size:self.size + count] = value[index:index + count]
                self.size += count

    def get(self, target):
        len1 = len(target)
        if self.size - self.position < len1:
            raise ValueError("get")
        index = 0
        for index in range(0, len1):
            target[index] = self._data[self.position]
            self.position = self.position + 1

    #
    #      Compares, whether two given arrays are similar starting from current
    #      position.
    #
    #      @param arr
    #                 Array to compare.
    # True, if arrays are similar.  False, if the arrays differ.
    #
    def compare(self, arr):
        len1 = len(arr)
        if not arr or (self.size - self.position < len1):
            return False
        bytes_ = bytearray(len1)
        self.get(bytes_)
        ret = arr == bytes_
        if not ret:
            self.position -= len1
        return ret

    #
    #      Reverses the order of the given array.
    #
    def reverse(self):
        first = self.position
        last = self.size - 1
        tmp = int()
        while last > first:
            tmp = self._data[last]
            self._data[last] = self._data[first]
            self._data[first] = tmp
            last -= 1
            first += 1

    #
    #      Push the given hex string as byte array into this buffer at the
    #      current
    #      position, and then increments the position.
    #
    #      @param value
    #                 Byte array to add.
    #      @param index
    #                 Byte index.
    #      @param count
    #                 Byte count.
    #
    def setHexString(self, value, index=0, count=None):
        tmp = self.hexToBytes(value)
        if count is None:
            count = len(tmp)
        self.set(tmp, index, count)

    def __str__(self):
        return self.hex(self._data, True, 0, self.size)

    #
    #      Get remaining data.
    #
    # Remaining data as byte array.
    #
    def remaining(self):
        return self.subArray(self.position, self.size - self.position)

    #
    #      Get remaining data as a hex string.
    #
    #      @param addSpace
    #                 Add space between bytes.
    # Remaining data as a hex string.
    #
    def remainingHexString(self, addSpace=True):
        return self.hex(self._data, addSpace, self.position, self.size - self.position)

    #
    #      Get data as hex string.
    #
    #      @param addSpace
    #                 Add space between bytes.
    #      @param index
    #                 Byte index.
    #      @param count
    #                 Byte count.
    # Data as hex string.
    #
    def toHex(self, addSpace=True, index=0, count=None):
        if count is None:
            count = len(self) - index
        return self.hex(self._data, addSpace, index, count)

    #Convert char hex value to byte value.
    @classmethod
    def ___getValue(cls, c):
        #Id char.
        if c.islower():
            c = c.upper()
        pos = GXByteBuffer.__HEX_ARRAY.find(c)
        if pos == -1:
            raise Exception("Invalid hex string")
        return pos

    @classmethod
    def hexToBytes(cls, value):
        """Convert string to byte array.
        value: Hex string.
        Returns byte array.
        """
        buff = bytearray()
        if value:
            lastValue = -1
            for ch in value:
                if ch != ' ':
                    if lastValue == -1:
                        lastValue = GXByteBuffer.___getValue(ch)
                    elif lastValue != -1:
                        buff.append(lastValue << GXByteBuffer.__NIBBLE | GXByteBuffer.___getValue(ch))
                        lastValue = -1
                elif lastValue != -1:
                    buff.append(GXByteBuffer.___getValue(ch))
                    lastValue = -1
        return buff

    @classmethod
    def hex(cls, value, addSpace=True, index=0, count=None):
        """
        Convert byte array to hex string.
        """
        #Return empty string if array is empty.
        if not value:
            return ""
        hexChars = ""
        #Python 2.7 handles bytes as a string array. It's changed to bytearray.
        if sys.version_info < (3, 0) and not isinstance(value, bytearray):
            value = bytearray(value)
        if count is None:
            count = len(value)
        for it in value[index:count]:
            hexChars += GXByteBuffer.__HEX_ARRAY[it >> GXByteBuffer.__NIBBLE]
            hexChars += GXByteBuffer.__HEX_ARRAY[it & GXByteBuffer.__LOW_BYTE_PART]
            if addSpace:
                hexChars += ' '
        return hexChars.strip()



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..enums import ErrorCode, DateTimeSkips
from ..internal._GXCommon import _GXCommon
from ..GXByteBuffer import GXByteBuffer
from ..GXDateTime import GXDateTime
from ..GXTime import GXTime
from ..GXDate import GXDate
from ..enums import ObjectType, DataType
from .enums import SingleActionScheduleType
from .GXDLMSScriptTable import GXDLMSScriptTable

# pylint: disable=too-many-instance-attributes
class GXDLMSActionSchedule(GXDLMSObject, IGXDLMSBase):
    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSActionSchedule
    """

    def __init__(self, ln="0.0.15.0.0.255", sn=0):
        """
        Constructor.

        ln : Logical Name of the object.
        sn : Short Name of the object.
        """
        GXDLMSObject.__init__(self, ObjectType.ACTION_SCHEDULE, ln, sn)
        self.type_ = SingleActionScheduleType.SingleActionScheduleType1
        # Script to execute.
        self.target = None
        self.executedScriptSelector = 0
        self.executionTime = list()

    def getValues(self):
        if self.target:
            ln = self.target.logicalName
        else:
            ln = ""
        return [self.logicalName,
                [ln, self.executedScriptSelector],
                self.type_,
                self.executionTime]

    #
    # Returns collection of attributes to read.  If attribute is static
    #      and
    # already read or device is returned HW error it is not returned.
    #
    def getAttributeIndexToRead(self, all_):
        attributes = list()
        #  LN is static and read only once.
        if all_ or not self.logicalName:
            attributes.append(1)
        #  ExecutedScriptLogicalName is static and read only once.
        if all_ or not self.isRead(2):
            attributes.append(2)
        #  Type is static and read only once.
        if all_ or not self.isRead(3):
            attributes.append(3)
        #  ExecutionTime is static and read only once.
        if all_ or not self.isRead(4):
            attributes.append(4)
        return attributes

    #
    # Returns amount of attributes.
    #
    def getAttributeCount(self):
        return 4

    #
    # Returns amount of methods.
    #
    def getMethodCount(self):
        return 0

    def getDataType(self, index):
        if index == 1:
            ret = DataType.OCTET_STRING
        elif index == 2:
            ret = DataType.ARRAY
        elif index == 3:
            return DataType.ENUM
        elif index == 4:
            return DataType.ARRAY
        else:
            raise ValueError("getDataType failed. Invalid attribute index.")
        return ret
    #
    # Returns value of given attribute.
    #
    def getValue(self, settings, e):
        if e.index == 1:
            return _GXCommon.logicalNameToBytes(self.logicalName)
        if e.index == 2:
            bb = GXByteBuffer()
            bb.setUInt8(DataType.STRUCTURE)
            bb.setUInt8(2)
            _GXCommon.setData(settings, bb, DataType.OCTET_STRING, _GXCommon.logicalNameToBytes(self.target.logicalName))
            _GXCommon.setData(settings, bb, DataType.UINT16, int(self.executedScriptSelector))
            return bb.array()
        if e.index == 3:
            return self.type_
        if e.index == 4:
            bb = GXByteBuffer()
            bb.setUInt8(DataType.ARRAY)
            if not self.executionTime:
                _GXCommon.setObjectCount(0, bb)
            else:
                _GXCommon.setObjectCount(len(self.executionTime), bb)
                for it in self.executionTime:
                    bb.setUInt8(DataType.STRUCTURE)
                    #  Count
                    bb.setUInt8(2)
                    #  Time
                    _GXCommon.setData(settings, bb, DataType.OCTET_STRING, GXTime(it))
                    #  Date
                    _GXCommon.setData(settings, bb, DataType.OCTET_STRING, GXDate(it))
            return bb.array()
        e.error = ErrorCode.READ_WRITE_DENIED
        return None

    #
    # Set value of given attribute.
    #
    def setValue(self, settings, e):
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            if e.value:
                ln = _GXCommon.toLogicalName(e.value[0])
                self.target = settings.objects.findByLN(ObjectType.SCRIPT_TABLE, ln)
                if not self.target:
                    self.target = GXDLMSScriptTable(ln)
                self.executedScriptSelector = e.value[1]
            else:
                self.target = None
                self.executedScriptSelector = 0
        elif e.index == 3:
            self.type_ = e.value
        elif e.index == 4:
            self.executionTime = []
            if e.value:
                for it in e.value:
                    time = GXDateTime(_GXCommon.changeType(settings, it[0], DataType.TIME))
                    date = GXDateTime(_GXCommon.changeType(settings, it[1], DataType.DATE))
                    tmp = GXDateTime(date)
                    tmp.value = tmp.value.replace(hour=time.value.hour, minute=time.value.minute, second=time.value.second)
                    tmp.skip = (date.skip & (DateTimeSkips.YEAR | DateTimeSkips.MONTH | DateTimeSkips.DAY | DateTimeSkips.DAY_OF_WEEK))
                    tmp.skip |= (time.skip & (DateTimeSkips.HOUR | DateTimeSkips.MINUTE | DateTimeSkips.SECOND | DateTimeSkips.MILLISECOND))
                    self.executionTime.append(tmp)
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    def load(self, reader):
        ot = ObjectType(reader.readElementContentAsInt("ObjectType"))
        ln = reader.readElementContentAsString("LN")
        if ot != ObjectType.NONE and ln:
            self.target = reader.objects.findByLN(ot, ln)
            #  if object is not load yet.
            if not self.target:
                self.target = GXDLMSScriptTable(ln)
        self.executedScriptSelector = reader.readElementContentAsInt("ExecutedScriptSelector")
        self.type_ = reader.readElementContentAsInt("Type")
        self.executionTime = []
        if reader.isStartElement("ExecutionTime", True):
            while reader.isStartElement("Time", False):
                it = reader.readElementContentAsDateTime("Time")
                self.executionTime.append(it)
            reader.readEndElement("ExecutionTime")

    def save(self, writer):
        if self.target:
            writer.writeElementString("ObjectType", int(self.target.objectType))
            writer.writeElementString("LN", self.target.logicalName)
        writer.writeElementString("ExecutedScriptSelector", self.executedScriptSelector)
        writer.writeElementString("Type", int(self.type_))
        if self.executionTime:
            writer.writeStartElement("ExecutionTime")
            for it in self.executionTime:
                writer.writeElementString("Time", it)
            writer.writeEndElement()

    def postLoad(self, reader):
        #  Upload target after load.
        if self.target:
            t = reader.objects.findByLN(ObjectType.SCRIPT_TABLE, self.target.logicalName)
            if t and self.target != t:
                self.target = t
